# Shadow module for module "OpenGL.GLUT"

try:

    # Try to import the original module...
    from OpenGL.GLUT import *
    
except:

    # Create dummy symbols...

    GLUT_ACCUM = 4
    GLUT_ACTIVE_ALT = 4
    GLUT_ACTIVE_CTRL = 2
    GLUT_ACTIVE_SHIFT = 1
    GLUT_ALPHA = 8
    GLUT_API_VERSION = 3
    GLUT_BITMAP_8_BY_13 = 3
    GLUT_BITMAP_9_BY_15 = 2
    GLUT_BITMAP_HELVETICA_10 = 6
    GLUT_BITMAP_HELVETICA_12 = 7
    GLUT_BITMAP_HELVETICA_18 = 8
    GLUT_BITMAP_TIMES_ROMAN_10 = 4
    GLUT_BITMAP_TIMES_ROMAN_24 = 5
    GLUT_BLUE = 2
    GLUT_CURSOR_BOTTOM_LEFT_CORNER = 19
    GLUT_CURSOR_BOTTOM_RIGHT_CORNER = 18
    GLUT_CURSOR_BOTTOM_SIDE = 13
    GLUT_CURSOR_CROSSHAIR = 9
    GLUT_CURSOR_CYCLE = 5
    GLUT_CURSOR_DESTROY = 3
    GLUT_CURSOR_FULL_CROSSHAIR = 102
    GLUT_CURSOR_HELP = 4
    GLUT_CURSOR_INFO = 2
    GLUT_CURSOR_INHERIT = 100
    GLUT_CURSOR_LEFT_ARROW = 1
    GLUT_CURSOR_LEFT_RIGHT = 11
    GLUT_CURSOR_LEFT_SIDE = 14
    GLUT_CURSOR_NONE = 101
    GLUT_CURSOR_RIGHT_ARROW = 0
    GLUT_CURSOR_RIGHT_SIDE = 15
    GLUT_CURSOR_SPRAY = 6
    GLUT_CURSOR_TEXT = 8
    GLUT_CURSOR_TOP_LEFT_CORNER = 16
    GLUT_CURSOR_TOP_RIGHT_CORNER = 17
    GLUT_CURSOR_TOP_SIDE = 12
    GLUT_CURSOR_UP_DOWN = 10
    GLUT_CURSOR_WAIT = 7
    GLUT_DEPTH = 16
    GLUT_DEVICE_IGNORE_KEY_REPEAT = 610
    GLUT_DEVICE_KEY_REPEAT = 611
    GLUT_DISPLAY_MODE_POSSIBLE = 400
    GLUT_DOUBLE = 2
    GLUT_DOWN = 0
    GLUT_ELAPSED_TIME = 700
    GLUT_ENTERED = 1
    GLUT_FULLY_COVERED = 3
    GLUT_FULLY_RETAINED = 1
    GLUT_GAME_MODE_ACTIVE = 0
    GLUT_GAME_MODE_DISPLAY_CHANGED = 6
    GLUT_GAME_MODE_HEIGHT = 3
    GLUT_GAME_MODE_PIXEL_DEPTH = 4
    GLUT_GAME_MODE_POSSIBLE = 1
    GLUT_GAME_MODE_REFRESH_RATE = 5
    GLUT_GAME_MODE_WIDTH = 2
    GLUT_GREEN = 1
    GLUT_HAS_DIAL_AND_BUTTON_BOX = 603
    GLUT_HAS_JOYSTICK = 612
    GLUT_HAS_KEYBOARD = 600
    GLUT_HAS_MOUSE = 601
    GLUT_HAS_OVERLAY = 802
    GLUT_HAS_SPACEBALL = 602
    GLUT_HAS_TABLET = 604
    GLUT_HIDDEN = 0
    GLUT_INDEX = 1
    GLUT_INIT_DISPLAY_MODE = 504
    GLUT_INIT_WINDOW_HEIGHT = 503
    GLUT_INIT_WINDOW_WIDTH = 502
    GLUT_INIT_WINDOW_X = 500
    GLUT_INIT_WINDOW_Y = 501
    GLUT_JOYSTICK_AXES = 615
    GLUT_JOYSTICK_BUTTONS = 614
    GLUT_JOYSTICK_BUTTON_A = 1
    GLUT_JOYSTICK_BUTTON_B = 2
    GLUT_JOYSTICK_BUTTON_C = 4
    GLUT_JOYSTICK_BUTTON_D = 8
    GLUT_JOYSTICK_POLL_RATE = 616
    GLUT_KEY_DOWN = 103
    GLUT_KEY_END = 107
    GLUT_KEY_F1 = 1
    GLUT_KEY_F10 = 10
    GLUT_KEY_F11 = 11
    GLUT_KEY_F12 = 12
    GLUT_KEY_F2 = 2
    GLUT_KEY_F3 = 3
    GLUT_KEY_F4 = 4
    GLUT_KEY_F5 = 5
    GLUT_KEY_F6 = 6
    GLUT_KEY_F7 = 7
    GLUT_KEY_F8 = 8
    GLUT_KEY_F9 = 9
    GLUT_KEY_HOME = 106
    GLUT_KEY_INSERT = 108
    GLUT_KEY_LEFT = 100
    GLUT_KEY_PAGE_DOWN = 105
    GLUT_KEY_PAGE_UP = 104
    GLUT_KEY_REPEAT_DEFAULT = 2
    GLUT_KEY_REPEAT_OFF = 0
    GLUT_KEY_REPEAT_ON = 1
    GLUT_KEY_RIGHT = 102
    GLUT_KEY_UP = 101
    GLUT_LAYER_IN_USE = 801
    GLUT_LEFT = 0
    GLUT_LEFT_BUTTON = 0
    GLUT_LUMINANCE = 512
    GLUT_MENU_IN_USE = 1
    GLUT_MENU_NOT_IN_USE = 0
    GLUT_MENU_NUM_ITEMS = 300
    GLUT_MIDDLE_BUTTON = 1
    GLUT_MULTISAMPLE = 128
    GLUT_NORMAL = 0
    GLUT_NORMAL_DAMAGED = 804
    GLUT_NOT_VISIBLE = 0
    GLUT_NUM_BUTTON_BOX_BUTTONS = 607
    GLUT_NUM_DIALS = 608
    GLUT_NUM_MOUSE_BUTTONS = 605
    GLUT_NUM_SPACEBALL_BUTTONS = 606
    GLUT_NUM_TABLET_BUTTONS = 609
    GLUT_OVERLAY = 1
    GLUT_OVERLAY_DAMAGED = 805
    GLUT_OVERLAY_POSSIBLE = 800
    GLUT_OWNS_JOYSTICK = 613
    GLUT_PARTIALLY_RETAINED = 2
    GLUT_RED = 0
    GLUT_RGB = 0
    GLUT_RGBA = 0
    GLUT_RIGHT_BUTTON = 2
    GLUT_SCREEN_HEIGHT = 201
    GLUT_SCREEN_HEIGHT_MM = 203
    GLUT_SCREEN_WIDTH = 200
    GLUT_SCREEN_WIDTH_MM = 202
    GLUT_SINGLE = 0
    GLUT_STENCIL = 32
    GLUT_STEREO = 256
    GLUT_STROKE_MONO_ROMAN = 1
    GLUT_STROKE_ROMAN = 0
    GLUT_TRANSPARENT_INDEX = 803
    GLUT_UP = 1
    GLUT_VIDEO_RESIZE_HEIGHT = 909
    GLUT_VIDEO_RESIZE_HEIGHT_DELTA = 905
    GLUT_VIDEO_RESIZE_IN_USE = 901
    GLUT_VIDEO_RESIZE_POSSIBLE = 900
    GLUT_VIDEO_RESIZE_WIDTH = 908
    GLUT_VIDEO_RESIZE_WIDTH_DELTA = 904
    GLUT_VIDEO_RESIZE_X = 906
    GLUT_VIDEO_RESIZE_X_DELTA = 902
    GLUT_VIDEO_RESIZE_Y = 907
    GLUT_VIDEO_RESIZE_Y_DELTA = 903
    GLUT_VISIBLE = 1
    GLUT_WINDOW_ACCUM_ALPHA_SIZE = 114
    GLUT_WINDOW_ACCUM_BLUE_SIZE = 113
    GLUT_WINDOW_ACCUM_GREEN_SIZE = 112
    GLUT_WINDOW_ACCUM_RED_SIZE = 111
    GLUT_WINDOW_ALPHA_SIZE = 110
    GLUT_WINDOW_BLUE_SIZE = 109
    GLUT_WINDOW_BUFFER_SIZE = 104
    GLUT_WINDOW_COLORMAP_SIZE = 119
    GLUT_WINDOW_CURSOR = 122
    GLUT_WINDOW_DEPTH_SIZE = 106
    GLUT_WINDOW_DOUBLEBUFFER = 115
    GLUT_WINDOW_FORMAT_ID = 123
    GLUT_WINDOW_GREEN_SIZE = 108
    GLUT_WINDOW_HEIGHT = 103
    GLUT_WINDOW_NUM_CHILDREN = 118
    GLUT_WINDOW_NUM_SAMPLES = 120
    GLUT_WINDOW_PARENT = 117
    GLUT_WINDOW_RED_SIZE = 107
    GLUT_WINDOW_RGBA = 116
    GLUT_WINDOW_STENCIL_SIZE = 105
    GLUT_WINDOW_STEREO = 121
    GLUT_WINDOW_WIDTH = 102
    GLUT_WINDOW_X = 100
    GLUT_WINDOW_Y = 101
    GLUT_XLIB_IMPLEMENTATION = 13

    def glutAddMenuEntry(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutAddSubMenu(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutAttachMenu(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutBitmapCharacter(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutBitmapLength(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutBitmapWidth(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutButtonBoxFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutChangeToMenuEntry(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutChangeToSubMenu(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutCopyColormap(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutCreateMenu(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutCreateSubWindow(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutCreateWindow(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutDestroyMenu(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutDestroyWindow(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutDetachMenu(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutDeviceGet(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutDialsFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutDisplayFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutEnterGameMode(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutEntryFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutEstablishOverlay(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutExtensionSupported(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutForceJoystickFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutFullScreen(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutGameModeGet(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutGameModeString(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutGet(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutGetColor(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutGetMenu(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutGetModifiers(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutGetWindow(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutHideOverlay(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutHideWindow(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutIconifyWindow(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutIdleFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutIgnoreKeyRepeat(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutInit(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutInitDisplayMode(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutInitDisplayString(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutInitWindowPosition(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutInitWindowSize(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutJoystickFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutKeyboardFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutKeyboardUpFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutLayerGet(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutLeaveGameMode(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutMainLoop(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutMenuStateFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutMenuStatusFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutMotionFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutMouseFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutOverlayDisplayFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutPassiveMotionFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutPopWindow(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutPositionWindow(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutPostOverlayRedisplay(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutPostRedisplay(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutPostWindowOverlayRedisplay(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutPostWindowRedisplay(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutPushWindow(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutRemoveMenuItem(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutRemoveOverlay(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutReportErrors(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutReshapeFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutReshapeWindow(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSetColor(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSetCursor(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSetIconTitle(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSetKeyRepeat(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSetMenu(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSetWindow(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSetWindowTitle(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSetupVideoResizing(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutShowOverlay(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutShowWindow(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSolidCone(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSolidCube(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSolidDodecahedron(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSolidIcosahedron(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSolidOctahedron(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSolidSphere(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSolidTeapot(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSolidTetrahedron(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSolidTorus(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSpaceballButtonFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSpaceballMotionFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSpaceballRotateFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSpecialFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSpecialUpFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutStopVideoResizing(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutStrokeCharacter(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutStrokeLength(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutStrokeWidth(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutSwapBuffers(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutTabletButtonFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutTabletMotionFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutTimerFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutUseLayer(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutVideoPan(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutVideoResize(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutVideoResizeGet(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutVisibilityFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutWarpPointer(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutWindowStatusFunc(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutWireCone(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutWireCube(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutWireDodecahedron(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutWireIcosahedron(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutWireOctahedron(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutWireSphere(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutWireTeapot(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutWireTetrahedron(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")

    def glutWireTorus(*args, **keyargs):
        raise ImportError("No module named OpenGL.GLUT. Please install PyOpenGL (http://pyopengl.sourceforge.net/).")




# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: mainwin.py,v 1.2 2006/01/27 07:52:40 mbaas Exp $

import wx
from . import window
#from cgkit import getApp
from .menu import *
from .panels import Panels, LayoutNode, PanelNode, PanelWidget, HORIZONTAL, VERTICAL
from .toolbars import ToolBarPalette

# MainWin
class MainWindow(window.Window):
    def __init__(self):
        window.Window.__init__(self)

        # Create menu
        file = Menu("&File", name="file", items=
                    [("&New"),
                     ("&Open..."),
                     ("&Save"),
                     ("Save &as...", None),
                     ("&Close"),
                     MenuItem("---",name="sep1"),
                     ("&Exit", None)
                     ])

        self.menu = Menu(items=[file])


        # ToolBar palette
        self.toolbars = ToolBarPalette(self.wx, self._menu_id_manager)

        # Create panels
        self.panels = Panels(self.wx)
        self.panels.layout = PanelNode(name="Dummy")


        self.mainlayout = wx.BoxSizer(wx.VERTICAL)
        self.mainlayout.AddSizer(self.toolbars.sizer, 0, wx.EXPAND, 0)
        self.mainlayout.Add(self.panels.wx, 1, wx.EXPAND, 0)
        self.wx.SetSizer(self.mainlayout)


#        views = LayoutNode(splittertype=VERTICAL, name="splitter1")

#        self.btn = wx.Button(parent, -1, "Button", wx.Point(0,0))
#        front = PanelNode(name="front", activatable=True,
#                          widget=PanelWidget(wx=self.btn))

#        dict = globals()
#        dict["app"]=getApp()
#        self.shell = wx.py.shell.Shell(parent, -1, locals=dict)
#        shell = PanelNode(name="shell", widget=PanelWidget(wx=self.shell))
        
#        views.setChildren((front, shell))
#        self.panels.layout = views
#        front.makeCurrent()

#        self.panels.updateLayout()

        if "mainwin.geometry" in getApp().prefs:
            x,y,w,h = getApp().prefs["mainwin.geometry"]
            print(("set",x,y,w,h))
            self.SetDimensions(x,y,w,h)
            if getApp().prefs["mainwin.maximized"]:
                self.Maximize(True)
           
        wx.EVT_SIZE(self, self.onResize)
        wx.EVT_MOVE(self, self.onMove)
        return

    def onMove(self, event):
        app = getApp()
        if "mainwin.geometry" in app.prefs:
            x,y,w,h = app.prefs["mainwin.geometry"]
        else:
            w,h = self.GetSizeTuple()
        x,y = self.GetPosition()
        if not self.IsMaximized():
            app.prefs["mainwin.geometry"] = [x,y,w,h]

    def onResize(self, event):
        app = getApp()
        if "mainwin.geometry" in app.prefs:
            x,y,w,h = app.prefs["mainwin.geometry"]
        else:
            x,y = self.GetPositionTuple()
        w,h = self.GetSize()
        if self.IsMaximized():
            app.prefs["mainwin.maximized"] = True
        else:
            app.prefs["mainwin.maximized"] = False
            app.prefs["mainwin.geometry"] = [x,y,w,h]
        event.Skip()




# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2009
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

import os, os.path, glob


class JobHandle(object):
    """Job object (queued).
    
    Objects of this class represent jobs that have already been queued.
    
    Every job has:
    
    - A type name
    - An instance name
    - A job description
    - An associated directory
    """
    
    def __init__(self, location, rootLocation):
        """Constructor.
        
        location is the path to the job directory (this may also point to a
        link file).
        rootLocation is the path to the job root directory (this is used
        for resolving job relative relative links).
        """
        object.__init__(self)
        # The job location (if useSymLinks is False, this may also point to a link file)
        self._location = str(location)

        # The job root directory
        self._rootLocation = rootLocation
        
        # The resolved location. This is always a directory (or technically, it
        # could be a non-existent path). If useSymLinks is True, this is
        # always the same as the above location.
        self._realLocation = self._resolveTxtLinks(self._location)
        
        
    def _resolveTxtLinks(self, location):
        """Resolve a path so that it points to a directory and not to a link file.
        
        If location already points to a directory, it is returned unmodified.
        If it points to a link file, the file is followed and the process
        repeats until a directory is encountered or the targegt does not exist.
        """
        visited = {}
        while os.path.exists(location) and os.path.isfile(location):
            visited[location] = True
            location = self._getTxtLinkTarget(location)
            if location in visited:
                raise ValueError("Cyclic directory links detected")
        return location
    
    def _getTxtLinkTarget(self, linkFileName):
        """Return the target that a link file points to.
        """
        f = open(linkFileName, "rt")
        # Read the first line which must be "[link]"
        s = f.readline()
        if s!="[link]\n":
            raise ValueError("%s is not a link file"%linkFileName)
        # Read the line containing the target
        targetLine = f.readline()
        f.close()
        # Check the target line
        if not targetLine.startswith("target="):
            raise ValueError("Error in link file %s"%linkFileName)
        target = targetLine[7:].strip()
        # The link must always be relative to the job root
        if not target.startswith("$JOBROOT"):
            raise ValueError("Invalid link target: %s"%target)
        target = self._rootLocation+target[8:]
        return target
        
    def __str__(self):
        if self.isFinished():
            state = "finished"
        elif self.isRunning():
            state = "running"
        elif self.isWaiting():
            state = "waiting"
        else:
            state = "?"
            
        res = "%s (%s)"%(os.path.basename(self._location), state)
        return res
        
    @property
    def location(self):
        """Return the job directory.
        """
        return self._location
    
    @property
    def label(self):
        """Return a short job label to quickly identify the job.
        """
        try:
            try:
                f = open(self.labelFile, "rt")
                label = f.read().strip()
            finally:
                f.close()
        except:
            label = ""

        if label=="":
            label = "Job %s"%self.number
        
        return label

    @property
    def number(self):
        """Return the job number.
        """
        nrStr = os.path.basename(self._location)[3:]
        try:
            return int(nrStr)
        except:
            return -1
        
    @property
    def submitTime(self):
        """Return the submission time (in seconds).
        
        May return None if the job doesn't exist or is broken.
        """
        try:
            s = os.stat(self.procDefFile)
            return s.st_mtime
        except OSError:
            return None

    @property
    def startTime(self):
        """Return the time the job was started (in seconds).
        
        Returns None if the job hasn't been started yet (or there is a problem
        with the job directory).
        """
        try:
            s = os.stat(self.pidFile)
            return s.st_mtime
        except OSError:
            return None

    @property
    def endTime(self):
        """Return the time the job was finished (in seconds).
        
        Returns None if the job hasn't been finished yet (or there is a problem
        with the job directory).
        """
        try:
            s = os.stat(self.finishedDir)
            return s.st_mtime
        except OSError:
            return None
        
    @property
    def progress(self):
        """Return the progress percentage value as an int.
        
        Returns None if the value couldn't be obtained for some reason.
        """
        try:
            progressFile = self.progressFile
            if os.path.exists(progressFile):
                res = os.path.getsize(self.progressFile)
                if res>100:
                    res = 100
            else:
                res = 0
            return res
        except:
            return None
    
    @property
    def statusLine(self):
        """Return a string containing the current status line for the GUI.
        
        The status line indicates what a running job is currently doing.
        """
        try:
            try:
                f = open(self.statusLineFile, "rt")
                line = f.read().strip()
            finally:
                f.close()
        except:
            line = ""

        return line
        

    @property
    def finishedDir(self):
        return os.path.join(self._realLocation, ".finished")
    
    @property
    def runningDir(self):
        return os.path.join(self._realLocation, ".running")

    @property
    def procDefFile(self):
        return os.path.join(self._realLocation, ".proc_def")

    @property
    def labelFile(self):
        """Return the location of the file that contains the job label.
        """
        return os.path.join(self._realLocation, ".label")

    @property
    def pidFile(self):
        """Return the location of the PID file.
        
        This file contains the PID of the process that is/was running the job.
        """
        return os.path.join(self.runningDir, ".pid")

    @property
    def hostFile(self):
        """Return the location of the host file.
        
        The host file contains the name of the host that is/was running the job.
        """
        return os.path.join(self.runningDir, ".host")

    @property
    def progressFile(self):
        """Return the location of the progress indicator file.
        """
        return os.path.join(self.runningDir, ".progress")

    @property
    def statusLineFile(self):
        """Return the location of the status line file.
        """
        return os.path.join(self.runningDir, ".statusline")

    @property
    def errorMarkerFile(self):
        """Return the location of the error marker file.
        
        The presence of this file indicates that running the job resulted
        in an error.
        """
        return os.path.join(self.runningDir, ".error_marker")

    @property
    def procTracebackFile(self):
        """Return the location of the file that contains the job procedure traceback.
        """
        return os.path.join(self.runningDir, ".proc_traceback")

    @property
    def stdoutFile(self):
        """Return the location of the file that contains stdout.
        """
        return os.path.join(self.runningDir, ".stdout")

    @property
    def stderrFile(self):
        """Return the location of the file that contains stderr.
        """
        return os.path.join(self.runningDir, ".stderr")
    
    def isWaiting(self):
        """Check if this job is currently in 'waiting' state.
        """
        if os.path.exists(self.finishedDir):
            return False
        if os.path.exists(self.runningDir):
            return False
        return os.path.exists(self.procDefFile)
    
    def isRunning(self):
        """Check if this job is currently in 'running' state.
        """
        if self.isFinished():
            return False
        else:
            return os.path.exists(self.runningDir)
    
    def isFinished(self):
        """Check if this job is currently in 'finished' state.
        """
        return os.path.exists(self.finishedDir)
    
    def listSubJobs(self):
        """Return a list of sub-jobs.
        
        The return value is a list of JobHandle objects.
        """
        subDirs = glob.glob(os.path.join(self._realLocation, "job*"))
        #subDirs = filter(lambda p: os.path.isdir(p), subDirs)
        jobs = [(jobDir, int(os.path.basename(jobDir)[3:])) for jobDir in subDirs]
        jobs.sort(key=lambda a: a[1])
        return [JobHandle(jobDir, self._rootLocation) for jobDir,nr in jobs]
    
    def hasError(self, recursive=False):
        """Check if this job produced an error.
        
        If recursive is False, only this job is considered, not the children
        jobs. In this case, the return value is only meaningful when the job
        is in finished state.
        If recursive is True, the result will be True if any job in the entire
        sub-hierarchy has an error. This can be an expensive operation because
        all sub-directories have to be checked.
        """
        err = os.path.exists(self.errorMarkerFile)
        if err:
            return True
        
        if recursive:
            subJobs = self.listSubJobs()
            for j in subJobs:
                if j.hasError(recursive=True):
                    return True
            
        return False
    
    def setError(self):
        """Set the error flag.
        
        This method may only be called by the process that is currently
        running the job (the .running directory must exist).
        """
        errFile = self.errorMarkerFile
        if not os.path.exists(errFile):
            f = open(errFile, "wb")
            f.close()



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: camcontrol.py,v 1.4 2005/08/31 14:49:13 mbaas Exp $

## \file camcontrol.py
## Contains the CameraControl class.

from . import component
from . import eventmanager
from .globalscene import getScene
from .events import *
from .keydefs import *
from .cgtypes import *

# CameraControl
class CameraControl(component.Component):
    """Camera control component.

    This component lets you control a camera interactively with the mouse.

    Todo: - About what point rotates the FreeCam?
          - Make movement distances customizable (auto configurable?)
    """
    
    def __init__(self,
                 cam,
                 name = "CameraControl",
                 mode = 0,
                 auto_insert = True):
        """Constructor.

        \param cam (\c ...Cam?...) Camera to control
        \param name (\c str) Name of the component
        \param mode (\c int) 0=Emulate MAX behavior, 1=Emulate Maya behavior
        \param auto_insert (\c bool) Insert into scene or not
        """
        
        component.Component.__init__(self, name=name, auto_insert=auto_insert)
        
        # mode: 0=Emulate MAX / 1=Emulate Maya / 2=Emulate Softimage (Navigation tool)
        self.mode = mode  
        self.cam = cam
        self.left_down = False
        self.middle_down = False
        self.right_down = False
        self.alt_down = False
        self.ctrl_down = False
        em = eventmanager.eventManager()
        em.connect(KEY_PRESS, self)
        em.connect(KEY_RELEASE, self)
        em.connect(LEFT_DOWN, self)
        em.connect(LEFT_UP, self)
        em.connect(MIDDLE_DOWN, self)
        em.connect(MIDDLE_UP, self)
        em.connect(RIGHT_DOWN, self)
        em.connect(RIGHT_UP, self)
        em.connect(MOUSE_MOVE, self)
        em.connect(MOUSE_WHEEL, self)
        em.connect(SPACE_MOTION, self)

    def panCondition(self):
        """Return True if the panning mode is active.
        """
        if self.mode==0:
            # MAX
            return self.middle_down and not self.alt_down and not self.ctrl_down
        elif self.mode==1:
            # Maya
            return self.middle_down and self.alt_down
        else:
            # Softimage
            return self.left_down

    def rotCondition(self):
        """Return True if rotation mode is active.
        """
        if self.mode==0:
            # MAX
            return self.middle_down and self.alt_down and not self.ctrl_down
        elif self.mode==1:
            # Maya
            return self.left_down and self.alt_down
        else:
            # Softimage
            return self.right_down

    def fbCondition(self):
        """Return True if 'forward/backward' movement mode is active.
        """        
        if self.mode==0:
            # MAX
            return self.middle_down and self.alt_down and self.ctrl_down
        elif self.mode==1:
            # Maya
            return self.right_down and self.alt_down
        else:
            # Softimage
            return self.middle_down

    def translateXYAmount(self, evt):
        """Return the amount of sideways movement."""
        if self.mode==0:
            # MAX
            return 0.003*evt.dx, 0.003*evt.dy
        elif self.mode==1:
            # Maya
            return 0.003*evt.dx, 0.003*evt.dy
        else:
            # Softimage
            return 0.002*evt.dx, 0.002*evt.dy

    def translateZAmount(self, evt):
        """Return the amount of 'forward/backward' movement."""
        if self.mode==0:
            # MAX
            return -0.01*evt.dy
        elif self.mode==1:
            # Maya
            return 0.004*(evt.dy+evt.dx)
        else:
            # Softimage
            return 0.008*(evt.dy-evt.dx)

    def rotationAmount(self, evt):
        """Return the amount of rotation."""
        if self.mode==0:
            # MAX
            return 0.01*evt.dx, 0.01*evt.dy
        elif self.mode==1:
            # Maya
            return 0.01*evt.dx, 0.01*evt.dy
        else:
            # Softimage
            return 0.015*evt.dx, 0.015*evt.dy

    # onKeyPress
    def onKeyPress(self, e):
        """Set key flags."""
        if e.altKey():
            self.alt_down = True
        elif e.controlKey():
            self.ctrl_down = True

    # onKeyRelease
    def onKeyRelease(self, e):
        """Unset key flags."""
        if e.altKey():
            self.alt_down = False
        elif e.controlKey():
            self.ctrl_down = False

    # onMouseMove
    def onMouseMove(self, e):

        if self.panCondition():
            dx, dy = self.translateXYAmount(e)
            self.translateXY(dx, dy)

        elif self.rotCondition():
            dx, dy = self.rotationAmount(e)
            self.rotate(dx, dy)
            
        elif self.fbCondition():
            f = self.translateZAmount(e)
            self.translateZ(f)

    def onMouseWheel(self, e):
        self.translateZ(0.002*e.delta)
                            
    def onLeftDown(self, e):
        self.left_down = True

    def onLeftUp(self, e):
        self.left_down = False

    def onMiddleDown(self, e):
        self.middle_down = True

    def onMiddleUp(self, e):
        self.middle_down = False

    def onRightDown(self, e):
        self.right_down = True

    def onRightUp(self, e):
        self.right_down = False

    # translateXY
    def translateXY(self, dx, dy):
        """Translate the camera in local XY.
        """

        T = self.cam.transform
        bx = vec3(tuple(T.getColumn(0))[:3])
        by = vec3(tuple(T.getColumn(1))[:3])
        if hasattr(self.cam, "target"):
            f = (self.cam.target-self.cam.pos).length()
            dx *= f
            dy *= f
        delta = dx*bx + dy*by
        self.cam.pos += delta
        if hasattr(self.cam, "target"):
            self.cam.target += delta

    # rotate
    def rotate(self, dx, dy):
        """Rotate around target."""
       
        T = self.cam.transform
        if hasattr(self.cam, "target"):
            pivot = self.cam.target
        else:
            pivot = vec3(0)
        up = getScene().up
        R = mat4(1).rotation(-dx, up)
        dp = self.cam.pos-pivot
        self.cam.pos = pivot + R*dp
            
        T = self.cam.transform
        bx = vec3(tuple(T.getColumn(0))[:3])
        R = mat4(1).rotation(dy, bx)
        dp = self.cam.pos-pivot
        self.cam.pos = pivot + R*dp

    # translateZ
    def translateZ(self, dz):
        """Move in viewing direction.
        """

        T = self.cam.transform
        bz = vec3(tuple(T.getColumn(2))[:3])
        if hasattr(self.cam, "target"):
            f = (self.cam.target-self.cam.pos).length()
        else:
            f = 1
        delta = f*dz*bz
        self.cam.pos += delta
#        self.cam.target += delta            
        

    # onSpaceMotion
    def onSpaceMotion(self, e):
        """
        this is still somewhat experimental
        (should distinguish between the camera types as it's meant to be
        used with a FreeCamera)
        """

        scene = getScene()

        tr = 0.001*e.translation
        rot = e.rotation

        if scene.handedness=='r':
            tr.x = -tr.x
            rot.y = -rot.y
            rot.z = -rot.z

        M = mat4().translation(tr)

        if rot!=vec3(0):
            a = 0.0002*rot.length()
            M = M*mat4().rotation(a, rot)

        T = self.cam.transform*M
        # Align camera to up direction
        R = T.getMat3()
        bz = vec3(tuple(T.getColumn(2))[:3])
        by = scene.up
        bx = by.cross(bz)
        by = bz.cross(bx)
        try:
            R = mat3(bx.normalize(), by.normalize(), bz.normalize())
        except:
            R = mat3(1)
        T.setMat3(R)

        self.cam.transform = T



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: glslangparams.py,v 1.2 2006/04/12 11:52:09 mbaas Exp $

"""Extract shader parameters from an OpenGL shader source file.
"""

import sys, copy
import io
from . import glslangtokenize
from . import simplecpp
from . import slparams
from .glslangtokenize import WHITESPACE, NAME, NUMBER, STRING, NEWLINE, OPERATOR, CHARACTER, TYPE, QUALIFIER

class GLSLangParseError(Exception):
    pass

# _GLSLangParser
class _GLSLangParser:
    """Extract variables from a glslang shader.
    """
    
    def __init__(self, file, structs=None):
        """Constructor.

        fils is a file-like object.
        structs is a dictionary containing the 'predefined' structs.
        This is used internally when the contents of a struct is
        recursively parsed.
        """

        # The following lists receive the result:

        # Contains 2-tuples (type, identifier)
        self.attribute = []
        # Contains 3-tuples (type, identifier, arraysize)
        self.varying = []
        # Contains 5-tuples (type, identifier, arraysize, structname, struct)
        self.uniform = []
        # Contains 5-tuples (type, identifier, arraysize, structname, struct)
        self.const = []
        # Contains 5-tuples (type, identifier, arraysize, structname, struct)
        self.other = []

        # file
        self.file = file
        # Current state
        self.state = self.initialState
        
        # Current qualifier
        self.qualifier = None
        # Current type
        self.type = None
        # Identifier
        self.identifier = None
        # String with the array size
        self.arraysize = None
        # Struct name
        self.structname = None
        self.struct = None

        self.named_structs = {}
        if structs!=None:
            self.named_structs = copy.deepcopy(structs)

        # Number of open parentheses
        self.openparen = 0

        # The contents of a struct as a string
        # (for calling the parser again on this)
        self.structcontents = ""
        

    def read(self):
        """Read the file.
        """
        glslangtokenize.tokenize(self.file.readline, self.tokeater)

    def tokeater(self, type, s, start, end, line, filename):
        """Token eater.

        This method skips whitespace and newlines and calls the
        current state method.
        """
        if type in [WHITESPACE, NEWLINE]:
            return

        self.state(type, s, start, end, line, filename)

    def switchState(self, state):
        """Switch to a different state.

        state is a callable.
        """
        self.state = state

    def variable(self):
        """Store a result variable.

        This method is called by the states whenever all information for
        a variable has been read.
        """
        if self.qualifier=="attribute":
            self.attribute.append((self.type, self.identifier))
        elif self.qualifier=="varying":
            self.varying.append((self.type, self.identifier, self.arraysize))
        elif self.qualifier=="uniform":
            self.uniform.append((self.type, self.identifier, self.arraysize, self.structname, self.struct))
        elif self.qualifier=="const":
            self.const.append((self.type, self.identifier, self.arraysize, self.structname, self.struct))
        else:
            self.other.append((self.type, self.identifier, self.arraysize, self.structname, self.struct))

    # States:
    # Each state takes the same arguments as a token reader

    def initialState(self, type, s, start, end, line, filename):
        """Initial state.
        """
        self.qualifier = None
        self.type = None
        self.identifier = None
        self.arraysize = None
        self.structname = None
        self.struct = None
        if type==QUALIFIER:
            self.qualifier = s
            self.switchState(self.qualifierState)
        elif type==TYPE and s!="struct":
            self.type = s
            self.switchState(self.typeState)
        elif s=="struct":
            self.type = s
            self.switchState(self.structState)
        else:
            # Is the type a previously defined struct?
            if s in self.named_structs:
                self.type = "struct"
                self.structname = s
                self.struct = self.named_structs[s]
                self.switchState(self.typeState)
            else:
                raise GLSLangParseError("%s, line %d: Syntax error: %s"%(filename, start[0], s))

    def qualifierState(self, type, s, start, end, line, filename):
        """A qualifier has been read.
        """
        if type==TYPE and s!="struct":
            self.type = s
            if self.qualifier in ["attribute", "varying"] and s not in ["float", "vec2", "vec3", "vec4", "mat2", "mat3", "mat4"]:
                raise GLSLangParseError("%s, line %d: Invalid type for an %s variable: %s"%(filename, start[0], self.qualifier, s))
            self.switchState(self.typeState)
        elif s=="struct":
            self.type = s
            self.switchState(self.structState)
        else:
            # Is the type a previously defined struct?
            if s in self.named_structs:
                self.type = "struct"
                self.structname = s
                self.struct = self.named_structs[s]
                if self.qualifier in ["attribute", "varying"]:
                    raise GLSLangParseError("%s, line %d: %s variables cannot be declared as structs"%(filename, start[0], self.qualifier))
                self.switchState(self.typeState)
            else:
                raise GLSLangParseError("%s, line %d: Syntax error: %s"%(filename, start[0], s))

    def typeState(self, type, s, start, end, line, filename):
        """The type of a varibale has been read.
        """
        if type==NAME:
            self.identifier = s
            self.switchState(self.nameState)
        elif s==";":
            self.switchState(self.initialState)
        else:
            raise GLSLangParseError("%s, line %d: Syntax error: %s"%(filename, start[0], s))

    def nameState(self, type, s, start, end, line, filename):
        """The name of a variable/function has been read.
        """
        if s==",":
            self.variable()
            self.switchState(self.typeState)
        elif s==";":
            self.variable()
            self.switchState(self.initialState)
        elif s=="[":
            self.arraysize = ""
            if self.qualifier=="attribute":
                raise GLSLangParseError("%s, line %d: attribute variables cannot be declared as arrays"%(filename, start[0]))
            self.switchState(self.arrayState)
        elif s=="(":
            self.openparen = 1
            self.switchState(self.functionState)
        elif s=="=":
            self.switchState(self.initState)
        else:
            raise GLSLangParseError("%s, line %d: Syntax error: %s"%(filename, start[0], s))

    def initState(self, type, s, start, end, line, filename):
        """A '=' has been encountered.

        Skip the initializer.
        """
        if s==";":
            self.variable()
            self.switchState(self.initialState)

    def arrayState(self, type, s, start, end, line, filename):
        """A '[' has been encountered.
        """
        if s=="]":
            self.switchState(self.arrayState2)
        else:
            self.arraysize += s

    def arrayState2(self, type, s, start, end, line, filename):
        if s==";":
            self.variable()
            self.switchState(self.initialState)
        elif s==",":
            self.variable()
            self.arraysize = None
            self.switchState(self.typeState)
        else:
            raise GLSLangParseError("%s, line %d: Syntax error: %s"%(filename, start[0], s))

    def functionState(self, type, s, start, end, line, filename):
        """Skip function args.
        """
        if s=="(":
            self.openparen += 1
        elif s==")":
            self.openparen -= 1
            if self.openparen==0:
                self.switchState(self.functionState2)
            
    def functionState2(self, type, s, start, end, line, filename):
        """Skip function body.
        """

        # If this was only a function prototype there will be an immediate
        # semicolon instead of the function body
        if self.openparen==0 and s==";":
            self.switchState(self.initialState)
        elif s=="{":
            self.openparen += 1
        elif s=="}":
            if self.openparen==0:
                raise GLSLangParseError("%s, line %d: '{' expected, got '}'"%(filename, start[0]))
            self.openparen -= 1
            if self.openparen==0:
                self.switchState(self.initialState)
        elif self.openparen==0:
            raise GLSLangParseError("%s, line %d: Syntax error: %s"%(filename, start[0], s))

    def structState(self, type, s, start, end, line, filename):
        """The keyword 'struct' has been read.
        """
        self.structcontents = ""
        if type==NAME:
            self.structname = s
            self.switchState(self.structState2)
        elif s=="{":
            self.switchState(self.structState3)
        else:
            raise GLSLangParseError("%s, line %d: Syntax error: %s"%(filename, start[0], s))

    def structState2(self, type, s, start, end, line, filename):
        """The struct name has been read.
        """
        if s=="{":
            self.switchState(self.structState3)
        else:
            raise GLSLangParseError("%s, line %d: Syntax error: %s"%(filename, start[0], s))

    def structState3(self, type, s, start, end, line, filename):
        """Collect the struct contents.
        """
        if s=="}":
            f = io.StringIO(self.structcontents)
            p = _GLSLangParser(f, structs=self.named_structs)
            p.read()
            self.struct = p.other
            if self.structname!="":
                self.named_structs[self.structname] = self.struct
            self.switchState(self.typeState)
        else:
            self.structcontents += s+" "
        


# glslangparams
def glslangparams(shader=None, cpp=None, cpperrstream=sys.stderr):
    """Extracts the shader parameters from an OpenGL 2 shader source file.

    The argument *shader* is either the name of the shader source file
    or a file-like object that provides the shader sources.
    *cpp* determines how the shader source is preprocessed. It
    can either be a string containing the name of an external
    preprocessor tool (such as ``cpp``) that must take the file name as
    parameter and dump the preprocessed output to stdout or it can be
    a callable that takes *shader* and *cpperrstream* as input and returns
    the preprocessed sources as a string. If the external
    preprocessor does not produce any data a :exc:`PreprocessorNotFound`
    exception is thrown.
    The error stream of the preprocessor is written to the object
    that is specified by *cpperrstream* which must have a :meth:`write()`
    method. If *cpperrstream* is ``None``, the error stream is ignored.

    If *cpp* is ``None`` a simple internal preprocessor based on the
    :mod:`simplecpp` module is used.

    The function returns three lists (*uniform*, *attribute*, *varying*)
    that contain the variables with the corresponding qualifier.

    A uniform variable is a 5-tuple (*type*, *identifier*, *arraysize*,
    *structname*, *struct*). *arraysize* is a string containing the
    expression for the length of the array (i.e. the value between
    the square brackets). If the variable is no array, *arraysize* is ``None``.
    When the variable is a struct, *type* has the value ``'struct'``. In this
    case, the struct is given in *struct* (which is itself a list of
    variables as 5-tuples). If the struct has a name, this name is
    given in *structname*, otherwise *structname* is ``None``.

    An attribute variable is a 2-tuple (*type*, *identifier*) and a
    varying variable is a 3-tuple (*type*, *identifier*, *arraysize*) where
    *arraysize* is defined as in the uniform case.
    """

    # Run the preprocessor on the input file...
    
    if cpp==None:
        cpp = simplecpp.PreProcessor()
        
    glslangsrc = slparams.preprocess(cpp, shader, cpperrstream)
    f = io.StringIO(glslangsrc)

    # Extract the variables...
    parser = _GLSLangParser(f)
    parser.read()
    return parser.uniform, parser.attribute, parser.varying





# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2010
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

from ctypes import *
from . import findlib
from . import decls
from .decls import SwsContext

def swscale_version():
    """Return the libswscale library version.
    
    Returns a tuple (major,minor,micro) containing the three parts of the
    version number. 
    """
    v = _lib().swscale_version()
    major = v>>16
    minor = (v>>8)&0xff
    micro = v&0xff
    return (major,minor,micro)

def sws_getContext(srcW, srcH, srcFormat, dstW, dstH, dstFormat, flags):
    """Allocate a SwsContext object.
    
    Must be deallocated using sws_freeContext().
    Returns a :class:`SwsContext` object.
    """
    func = _lib().sws_getContext
    func.restype = POINTER(SwsContext)
    res = func(srcW, srcH, srcFormat, dstW, dstH, dstFormat, flags, 0, 0, 0)
    if res:
        return res.contents
    else:
        return None

def sws_freeContext(swsCtx):
    """Free a SwsContext object.
    """
    _lib().sws_freeContext(byref(swsCtx))   
    
def sws_scale(swsCtx, src, srcStride, srcSliceY, srcSliceH, dst, dstStride):
    """Scale/convert an image.
    
    Scales the image slice in srcSlice and puts the resulting scaled
    slice in the image in dst. A slice is a sequence of consecutive
    rows in an image.

    Slices have to be provided in sequential order, either in
    top-bottom or bottom-top order. If slices are provided in
    non-sequential order the behavior of the function is undefined.

    *swsCtx* is the scaling context as returned by sws_getContext().
    *src* is the array containing the pointers to the planes of the source slice.
    *srcStride* is the array containing the strides for each plane of the
    source image
    *srcSliceY* is the position in the source image of the slice to process,
    that is the number (counted starting from zero) in the image of the first
    row of the slice.
    *srcSliceH* is the height of the source slice, that is the number
    of rows in the slice.
    *dst* is the array containing the pointers to the planes of the destination
    image.
    *dstStride* is the array containing the strides for each plane of
    the destination image
    Returns the height of the output slice.
    """
    ret = _lib().sws_scale(byref(swsCtx), src, srcStride, srcSliceY, srcSliceH, dst, dstStride)
    if ret<0:
        raise RuntimeError("Error: %s"%ret)
    return ret


# By default, try to load the swscale library that has the same major version
# than the one that was used for creating the cppdefs and decls module.
_libname = "swscale.%s"%(decls.LIBSWSCALE_VERSION_MAJOR)
_libswscale = None

def _lib():
    """Return the swscale shared library.
    """
    global _libswscale
    global _libname
    if _libswscale is None:
        _libswscale = findlib.findFfmpegLib(_libname)
    return _libswscale

    


# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: expression.py,v 1.3 2005/10/18 07:02:39 mbaas Exp $

## \file expression.py
## Contains the Expression class.

"""This module contains the Expression class."""

from .Interfaces import *
from . import protocols
from . import slots
from .eventmanager import eventManager
from .globalscene import getScene
from . import events
from .cgtypes import *
from .component import *
from .sl import *
from math import *
from . import _core
       

# Expression
class Expression(Component):
    """Compute an output value using a user defined expression.

    This component outputs a single value that is driven by a user defined
    expression. The expression is specified by a string and can use an
    arbitrary number of parameters. The parameters and their default values
    have to be provided to the constructor via keyword arguments. An exception
    is the special variable "t" which will always hold the current time
    (unless you declare it explicitly).
    For each parameter a slot is created (<name>_slot), so it is also
    possible to animate the parameters. The output value can be accessed via
    the "output" and "output_slot" attributes.

    Example:

    \code
    s = Sphere()
    e = Expression("1.0 + amp*sin(freq*t)", amp=0.2, freq=2.0)
    e.output_slot.connect(s.radius_slot)
    \endcode
    
    """

    protocols.advise(instancesProvide=[ISceneItem])

    def __init__(self,
                 expr = "",
                 exprtype = None,
                 name = "Expression",
                 **keyargs
                 ):
        """Constructor.

        If no expression type is given the component tries to determine
        the type itself by executing the expression and inspecting the
        return type.

        \param expr (\c str) Expression
        \param exprtype (\c str) Output type or None
        \param name (\c str) Component name
        \param keyargs Parameters used in the expression
        """
        Component.__init__(self, name=name)

        self.expr = expr
        self.exprtype = exprtype

        # Create a parameter slot for every extra key arg...
        for k in keyargs:
            T = type(keyargs[k])
            if T==float or T==int:
                typ = "double"
                valstr = str(keyargs[k])
            elif T==vec3:
                typ = "vec3"
                x, y, z = keyargs[k]
                valstr = "vec3(%s, %s, %s)"%(x,y,z)
            elif T==vec4:
                typ = "vec4"
                x, y, z, w = keyargs[k]
                valstr = "vec4(%s, %s, %s, %s)"%(x,y,z,w)
            elif T==mat3:
                typ = "mat3"
                valstr = "mat3(%s)"%keyargs[k].toList(rowmajor=True)
            elif T==mat4:
                typ = "mat4"
                valstr = "mat4(%s)"%keyargs[k].toList(rowmajor=True)
            elif T==quat:
                typ = "quat"
                w, x, y, z = keyargs[k]
                valstr = "quat(%s, %s, %s, %s)"%(w,x,y,z)
            else:
                typ = "py"
                valstr = "keyargs[k]"
#                raise ValueError("Unsupported type: %s"%T)
            # Create slot
            exec("self.%s_slot = %sSlot(%s)"%(k, typ.capitalize(), valstr))
            exec("self.addSlot(k, self.%s_slot)"%k)

        # If t was not explicitly given use the timer...
        if "t" not in keyargs:
            self.t_slot = DoubleSlot()
            getScene().timer().time_slot.connect(self.t_slot)
            self.addSlot("t", self.t_slot)

        # Store a list of all parameter names
        self.vars = list(keyargs.keys())
        if "t" not in self.vars:
            self.vars.append("t")

        if self.exprtype==None:
            self.exprtype = self._determineReturnType()

        # Create the output slot
        e = self.exprtype
        if e.lower()=="float":
            e = "double"
        exec("self.output_slot = Procedural%sSlot(self.outProc)"%e.capitalize())
        self.addSlot("output", self.output_slot)

        # Create dependencies
        for v in self.vars:
            exec("self.%s_slot.addDependent(self.output_slot)"%(v))

            

    def protocols(self):
        return [ISceneItem, IComponent]

    def outProc(self):
        for _v in self.vars:
            exec("%s = self.%s_slot.getValue()"%(_v, _v))
        return eval("%s(%s)"%(self.exprtype, self.expr))
     
    ## protected:
        
    # "output" property...
    exec(slotPropertyCode("output"))

    def _determineReturnType(self):
        """Try to execute the stored expression and return the output type.
        """
        for _v in self.vars:
            exec("%s = self.%s_slot.getValue()"%(_v, _v))
        out = eval(self.expr)
        T = type(out)
        if T==float or T==int:
            return "float"
        if isinstance(out, _core.vec3):
            return "vec3"
        if isinstance(out, _core.vec4):
            return "vec4"
        if isinstance(out, _core.mat3):
            return "mat3"
        if isinstance(out, _core.mat4):
            return "mat4"
        if isinstance(out, _core.quat):
            return "quat"

        if T==tuple or T==list:
            if len(out)==3:
                return "vec3"
            if len(out)==4:
                return "vec4"
            if len(out)==9:
                return "mat3"
            if len(out)==16:
                return "mat4"
            raise ValueError("Unsupported sequence size: %d"%len(out))

        raise ValueError("Unknown expression type: %s"%T)
        
        
        

        




# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

"""Dummy module that is used as a replacement for PyProtocols for the time being.
"""

import inspect

# Key:Class name - Value:Dict (key:interface obj, value:None or adapter class)
_interfaces = {}

def adapt(obj, interface):
    """Dummy implementation.
    """
    global _interfaces
    
    className = obj.__class__.__name__
    interfaces = _interfaces.get(className, {})
    
#    print "ADAPT",obj.__class__.__name__, interface, interfaces.has_key(interface)
    if interface in interfaces:
        return obj
    else:
        raise NotImplementedError("interface not implemented")

def advise(instancesProvide, asAdapterForTypes=None):
    """Dummy implementation.
    """
    global _interfaces
    
    frame = inspect.currentframe()
    outer = inspect.getouterframes(frame)
    currentClassName = outer[1][3]

#    print "ADVISE",currentClassName, asAdapterForTypes
    
    for interface in instancesProvide:
        if currentClassName not in _interfaces:
            _interfaces[currentClassName] = {}
        _interfaces[currentClassName][interface] = 1
#        print "    %s implements %s"%(currentClassName, interface)

class Interface:
    """Dummy interface base class.
    """
    pass



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: lwob.py,v 1.1 2006/03/12 22:41:42 mbaas Exp $

import struct
from .cgtypes import vec3

class LWOBError(Exception):
    pass

# Texture
class Texture:
    """Texture description.

    A value remains None if it wasn't set in the LWOB file.
    """
    def __init__(self):
        self.type = None        # str
        self.flags = None       # int
        self.size = None        # vec3
        self.center = None      # vec3
        self.falloff = None     # vec3
        self.velocity = None    # vec3
        self.color = None       # 3-tuple
        self.value = None       # int
        self.bumpamplitude = None # float
        self.floatparams = 10*[None]  # list of floats
        self.intparams = 10*[None]    # list of ints
        self.imagename = None   # str
        self.alphaname = None   # str
        self.widthwrap = None   # int
        self.heightwrap = None  # int
        self.antialiasingstrength = None # float
        self.opacity = None     # float
        self.shader = []        # list of str
        self.shaderdata = []    # list of str
        self.seq_offset = None  # int
        self.seq_flags = None   # int
        self.seq_looplength = None # int
        self.flyer_begin = None # int
        self.flyer_end = None   # int
        self.cycle_speed = None # int
        self.cycle_low = None   # int
        self.cycle_high = None  # int

# Surface
class Surface:
    """Surface description.

    A value remains None if it wasn't set in the LWOB file.    
    """
    
    def __init__(self, name):
        self.name = name
        
        self.color = None          # 3-tuple
        self.flags = None          # int
        self.luminosity = None     # float
        self.diffuse = None        # float
        self.specular = None       # float
        self.reflection = None     # float
        self.transparency = None   # float
        self.glossiness = None     # int
        self.reflectionmode = None # int
        self.refmap = None         # str
        self.refmap_seamangle = None # float
        self.refractiveindex = None  # float
        self.edgetransp = None     # float
        self.maxsmoothangle = None # float

        self.tex_color = []        # list of Texture objects
        self.tex_diffuse = []
        self.tex_specular = []
        self.tex_reflection = []
        self.tex_transparency = []
        self.tex_luminosity = []
        self.tex_bump = []
        

# LWOBReader
class LWOBReader:
    """Lightwave Object reader.

    This class reads Lightwave Object (*.lwo) files and calls handler
    methods for each chunk found. The following handler methods can be
    implemented in derived classes:

    - handlePNTS(points):  points is a list of vec3 objects containing
                           all vertices.
    - handleSRFS(names):   names is a list of strings containing the names
                           of all surface definitions.
    - handlePOLS(polys):   polys is a list of polygons. Each polygon is a
                           tuple (verts, surface) where verts is a list of
                           vertex indices and surface the index of the surface
                           that is used by this polygon.
    - handleCRVS(curves):  curves is a list of curves. Each curve is a
                           tuple (verts, surface, flags).
    - handlePCHS(patches): patches is a list of patches. Each patch is a
                           tuple (verts, surface).
    - handleSURF(surface): surface is a Surface object that contains the
                           surface attributes.
    """

    def __init__(self):
        """Constructor.
        """
        # The file handle
        self._file = None
        # The number of bytes that are left in the file.
        # After the FORM has been read, it is initialized with the
        # FORM length and is used to determine whether there are still
        # more chunks left or not.
        self._bytes_left = 12

        # Current Surface object when reading a SURF chunk
        self._current_surface = None
        # Current Texture object
        self._current_tex = None

    def read(self, file):
        """Read the lwo file.

        file is a file-like object that provides the read() and seek()
        methods.
        """
        self._file = file
        self._bytes_left = 12
        
        # Read the FORM header...
        tag, length = self.readChunkHeader()
        self.onChunk(tag, length, 0)
        self._bytes_left = length
        if tag!="FORM":
            raise LWOBError("Not a Lightwave object file")

        # Read the LWOB tag...
        lwob = self.readNBytes(4)
        if lwob!="LWOB":
            raise LWOBError("Not a Lightwave object file")

        # Read the contained chunks and call the 'reader' method if available.
        # If no reader is defined for a chunk this chunk is skipped.
        while not self._eofReached():
            tag,length = self.readChunkHeader()
            self.onChunk(tag, length, 1)
#            print tag, length
            handlername = "read%s"%tag
            handler = getattr(self, handlername, None)
            if handler==None:
                self.skipChunk(length)
            else:
                handler(length)


    # Handler methods. These methods can be overridden in derived classes.

    def onChunk(self, tag, length, level): pass

    def handlePNTS(self, points): pass
    def handleSRFS(self, names): pass
    def handlePOLS(self, polys): pass
    def handleCRVS(self, curves): pass
    def handlePCHS(self, patches): pass
    def handleSURF(self, surface): pass


    ### readXXXX() reader methods (XXXX is the chunk tag)
    ### A reader takes the chunk length as argument and has to read
    ### its data from the file and call an appropriate handler method
    ### (handleXXXX(...))

    def readPNTS(self, length):
        """Read a PNTS chunk and call the PNTS handler.
        """
        if length%12!=0:
            raise LWOBError("Invalid PNTS chunk size (%d)"%length)
        
        numpnts = int(length/12)
        # Read the chunk data...
        data = self.readNBytes(length)
        # Unpack the coordinates...
        coords = struct.unpack(">%s"%(numpnts*"fff"), data)
        # Initialize vec3s...
        pnts = []
        for i in range(numpnts):
            pnts.append(vec3(coords[i*3:(i+1)*3]))
            
        self.handlePNTS(pnts)

    def readSRFS(self, length):
        """Read a SRFS chunk and call the SRFS handler.
        """
        data = self.readNBytes(length)
        names = []
        while len(data)>0:
            n = data.find("\000")
            if n==-1:
                n = len(data)
            names.append(data[:n])
            if n%2==0:
                n+=1
            data = data[n+1:]

        self.handleSRFS(names)

    def readPOLS(self, length):
        """Read a POLS chunk and call the POLS handler.
        """
        data = self.readNBytes(length)
        idx = 0
        unpack = struct.unpack
        polys = []
        # The number of detail polygons that follow
        # (detail polygons are ignored)
        details_left = 0
        while idx<len(data):
            numverts = unpack(">H", data[idx:idx+2])[0]
            idx += 2
            verts = unpack(">%s"%(numverts*"H"), data[idx:idx+numverts*2])
            idx += numverts*2
            surf = unpack(">H", data[idx:idx+2])[0]
            idx += 2
            if details_left==0:
                polys.append((verts, surf))
            else:
                details_left -= 1

            # Does the polygon have detail polygons?
            if surf<0:
                details_left = unpack(">H", data[idx:idx+2])[0]
                idx += 2

        self.handlePOLS(polys)

    def readCRVS(self, length):
        """Read a CRVS chunk and call the CRVS handler.
        """
        data = self.readNBytes(length)
        idx = 0
        unpack = struct.unpack
        curves = []
        while idx<len(data):
            numverts = unpack(">H", data[idx:idx+2])[0]
            idx += 2
            verts = unpack(">%s"%(numverts*"H"), data[idx:idx+numverts*2])
            idx += numverts*2
            surf,flags = unpack(">HH", data[idx:idx+4])
            idx += 4
            curves.append((verts, surf, flags))

        self.handleCRVS(curves)
        
    def readPCHS(self, length):
        """Read a PCHS chunk and call the PCHS handler.
        """
        data = self.readNBytes(length)
        idx = 0
        unpack = struct.unpack
        patches = []
        while idx<len(data):
            numverts = unpack(">H", data[idx:idx+2])[0]
            idx += 2
            verts = unpack(">%s"%(numverts*"H"), data[idx:idx+numverts*2])
            idx += numverts*2
            surf = unpack(">H", data[idx:idx+2])[0]
            idx += 2
            patched.append((verts, surf))

        self.handlePCHS(patches)

    def readSURF(self, length):
        """Read a SURF chunk and call the SURF handler.
        """
        # Read the surface name byte by byte...
        name = ""
        while 1:
            c = self.readNBytes(1)
            length -= 1
            if ord(c)==0:
                if len(name)%2==0:
                    # Skip the pad byte
                    self.readNBytes(1)
                    length -= 1
                break
            name += c

        self._current_surface = Surface(name)
        self._current_tex = None
        
#        print "SURFACE",name
        while length>0:
            tag,sublength = self.readSubChunkHeader()
            self.onChunk(tag, sublength, 2)
            length -= 6+sublength
#            print " ",tag,sublength
            handlername = "readSURF_%s"%tag
            handler = getattr(self, handlername, None)
            if handler==None:
                self.skipChunk(sublength)
            else:
                handler(sublength)

        self.handleSURF(self._current_surface)
        self._current_tex = None
        self._current_surface = None

    def readSURF_COLR(self, length):
        """COLR chunk."""
        if length!=4:
            raise LWOBError("Invalid COLR size (%d instead of 4)"%length)
        data = self.readNBytes(4)
        color = struct.unpack(">BBB", data[:3])
        self._current_surface.color = color

    def readSURF_FLAG(self, length):
        """FLAG chunk."""
        if length!=2:
            raise LWOBError("Invalid FLAG size (%d instead of 2)"%length)
        data = self.readNBytes(2)
        flags = struct.unpack(">H", data)[0]
        self._current_surface.flags = flags
        
    def readSURF_LUMI(self, length):
        """LUMI chunk."""
        if length!=2:
            raise LWOBError("Invalid LUMI size (%d instead of 2)"%length)
        data = self.readNBytes(2)
        lumi = struct.unpack(">H", data)[0]
        if self._current_surface.luminosity==None:
            self._current_surface.luminosity = lumi/256.0

    def readSURF_VLUM(self, length):
        """VLUM chunk."""
        if length!=4:
            raise LWOBError("Invalid VLUM size (%d instead of 4)"%length)
        data = self.readNBytes(4)
        lum = struct.unpack(">f", data)[0]
        self._current_surface.luminosity = lum

    def readSURF_DIFF(self, length):
        """DIFF chunk."""
        if length!=2:
            raise LWOBError("Invalid DIFF size (%d instead of 2)"%length)
        data = self.readNBytes(2)
        diff = struct.unpack(">H", data)[0]
        if self._current_surface.diffuse==None:
            self._current_surface.diffuse = diff/256.0

    def readSURF_VDIF(self, length):
        """VDIF chunk."""
        if length!=4:
            raise LWOBError("Invalid VDIF size (%d instead of 4)"%length)
        data = self.readNBytes(4)
        diff = struct.unpack(">f", data)[0]
        self._current_surface.diffuse = diff

    def readSURF_SPEC(self, length):
        """SPEC chunk."""
        if length!=2:
            raise LWOBError("Invalid SPEC size (%d instead of 2)"%length)
        data = self.readNBytes(2)
        spec = struct.unpack(">H", data)[0]
        if self._current_surface.specular==None:
            self._current_surface.specular = spec/256.0

    def readSURF_VSPC(self, length):
        """VSPC chunk."""
        if length!=4:
            raise LWOBError("Invalid VSPC size (%d instead of 4)"%length)
        data = self.readNBytes(4)
        spec = struct.unpack(">f", data)[0]
        self._current_surface.specular = spec

    def readSURF_REFL(self, length):
        """REFL chunk."""
        if length!=2:
            raise LWOBError("Invalid REFL size (%d instead of 2)"%length)
        data = self.readNBytes(2)
        refl = struct.unpack(">H", data)[0]
        if self._current_surface.reflection==None:
            self._current_surface.reflection = refl/256.0

    def readSURF_VRFL(self, length):
        """VRFL chunk."""
        if length!=4:
            raise LWOBError("Invalid VRFL size (%d instead of 4)"%length)
        data = self.readNBytes(4)
        refl = struct.unpack(">f", data)[0]
        self._current_surface.reflection = refl

    def readSURF_TRAN(self, length):
        """TRAN chunk."""
        if length!=2:
            raise LWOBError("Invalid TRAN size (%d instead of 2)"%length)
        data = self.readNBytes(2)
        tran = struct.unpack(">H", data)[0]
        if self._current_surface.transparency==None:
            self._current_surface.transparency = tran/256.0

    def readSURF_VTRN(self, length):
        """VTRN chunk."""
        if length!=4:
            raise LWOBError("Invalid VTRN size (%d instead of 4)"%length)
        data = self.readNBytes(4)
        tran = struct.unpack(">f", data)[0]
        self._current_surface.transparency = tran

    def readSURF_GLOS(self, length):
        """GLOS chunk."""
        if length!=2:
            raise LWOBError("Invalid GLOS size (%d instead of 2)"%length)
        data = self.readNBytes(2)
        glos = struct.unpack(">H", data)[0]
        self._current_surface.glossiness = glos

    def readSURF_RFLT(self, length):
        """RFLT chunk."""
        if length!=2:
            raise LWOBError("Invalid RFLT size (%d instead of 2)"%length)
        data = self.readNBytes(2)
        mode = struct.unpack(">H", data)[0]
        self._current_surface.reflectionmode = mode

    def readSURF_RIMG(self, length):
        """RIMG chunk."""
        data = self.readNBytes(length)
        n = data.find("\000")
        if n!=-1:
            data = data[:n]
        self._current_surface.refmap = data

    def readSURF_RSAN(self, length):
        """RSAN chunk."""
        if length!=4:
            raise LWOBError("Invalid RSAN size (%d instead of 4)"%length)
        data = self.readNBytes(4)
        deg = struct.unpack(">f", data)[0]
        self._current_surface.refmap_seamangle = deg

    def readSURF_RIND(self, length):
        """RIND chunk."""
        if length!=4:
            raise LWOBError("Invalid RIND size (%d instead of 4)"%length)
        data = self.readNBytes(4)
        ior = struct.unpack(">f", data)[0]
        self._current_surface.refractiveindex = ior

    def readSURF_EDGE(self, length):
        """EDGE chunk."""
        if length!=4:
            raise LWOBError("Invalid EDGE size (%d instead of 4)"%length)
        data = self.readNBytes(4)
        edge = struct.unpack(">f", data)[0]
        self._current_surface.edgetransp = edge

    def readSURF_SMAN(self, length):
        """SMAN chunk."""
        if length!=4:
            raise LWOBError("Invalid SMAN size (%d instead of 4)"%length)
        data = self.readNBytes(4)
        sman = struct.unpack(">f", data)[0]
        self._current_surface.maxsmoothangle = sman

    def readSURF_CTEX(self, length):
        """CTEX chunk."""
        data = self.readNBytes(length)
        n = data.find("\000")
        if n!=-1:
            data = data[:n]

        tex = Texture()
        tex.type = data
        self._current_tex = tex
        self._current_surface.tex_color.append(tex)

    def readSURF_DTEX(self, length):
        """DTEX chunk."""
        data = self.readNBytes(length)
        n = data.find("\000")
        if n!=-1:
            data = data[:n]

        tex = Texture()
        tex.type = data
        self._current_tex = tex
        self._current_surface.tex_diffuse.append(tex)

    def readSURF_STEX(self, length):
        """STEX chunk."""
        data = self.readNBytes(length)
        n = data.find("\000")
        if n!=-1:
            data = data[:n]

        tex = Texture()
        tex.type = data
        self._current_tex = tex
        self._current_surface.tex_specular.append(tex)

    def readSURF_RTEX(self, length):
        """RTEX chunk."""
        data = self.readNBytes(length)
        n = data.find("\000")
        if n!=-1:
            data = data[:n]

        tex = Texture()
        tex.type = data
        self._current_tex = tex
        self._current_surface.tex_reflection.append(tex)

    def readSURF_TTEX(self, length):
        """TTEX chunk."""
        data = self.readNBytes(length)
        n = data.find("\000")
        if n!=-1:
            data = data[:n]

        tex = Texture()
        tex.type = data
        self._current_tex = tex
        self._current_surface.tex_transparency.append(tex)

    def readSURF_LTEX(self, length):
        """LTEX chunk."""
        data = self.readNBytes(length)
        n = data.find("\000")
        if n!=-1:
            data = data[:n]

        tex = Texture()
        tex.type = data
        self._current_tex = tex
        self._current_surface.tex_luminosity.append(tex)

    def readSURF_BTEX(self, length):
        """BTEX chunk."""
        data = self.readNBytes(length)
        n = data.find("\000")
        if n!=-1:
            data = data[:n]

        tex = Texture()
        tex.type = data
        self._current_tex = tex
        self._current_surface.tex_bump.append(tex)

    def readSURF_TFLG(self, length):
        """TFLG chunk."""
        if length!=2:
            raise LWOBError("Invalid TFLG size (%d instead of 2)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TFLG chunk")
        data = self.readNBytes(2)
        flags = struct.unpack(">H", data)[0]
        self._current_tex.flags = flags

    def readSURF_TSIZ(self, length):
        """TSIZ chunk."""
        if length!=12:
            raise LWOBError("Invalid TSIZ size (%d instead of 12)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TSIZ chunk")
        data = self.readNBytes(12)
        v = vec3(struct.unpack(">fff", data))
        self._current_tex.size = v

    def readSURF_TCTR(self, length):
        """TCTR chunk."""
        if length!=12:
            raise LWOBError("Invalid TCTR size (%d instead of 12)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TCTR chunk")
        data = self.readNBytes(12)
        v = vec3(struct.unpack(">fff", data))
        self._current_tex.center = v

    def readSURF_TFAL(self, length):
        """TFAL chunk."""
        if length!=12:
            raise LWOBError("Invalid TFAL size (%d instead of 12)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TFAL chunk")
        data = self.readNBytes(12)
        v = vec3(struct.unpack(">fff", data))
        self._current_tex.falloff = v

    def readSURF_TVEL(self, length):
        """TVEL chunk."""
        if length!=12:
            raise LWOBError("Invalid TVEL size (%d instead of 12)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TVEL chunk")
        data = self.readNBytes(12)
        v = vec3(struct.unpack(">fff", data))
        self._current_tex.velocity = v

    def readSURF_TCLR(self, length):
        """TCLR chunk."""
        if length!=4:
            raise LWOBError("Invalid TCLR size (%d instead of 4)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TCLR chunk")
        data = self.readNBytes(4)
        col = struct.unpack(">BBB", data[:3])
        self._current_tex.color = v

    def readSURF_TVAL(self, length):
        """TVAL chunk."""
        if length!=2:
            raise LWOBError("Invalid TVAL size (%d instead of 2)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TVAL chunk")
        data = self.readNBytes(2)
        val = struct.unpack(">H", data)[0]
        self._current_tex.value = val

    def readSURF_TAMP(self, length):
        """TAMP chunk."""
        if length!=4:
            raise LWOBError("Invalid TAMP size (%d instead of 4)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TAMP chunk")
        data = self.readNBytes(4)
        amp = struct.unpack(">f", data)[0]
        self._current_tex.bumpamplitude = amp

    def readSURF_TFP0(self, length):
        """TFP0 chunk."""
        if length!=4:
            raise LWOBError("Invalid TFP0 size (%d instead of 4)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TFP0 chunk")
        data = self.readNBytes(4)
        f = struct.unpack(">f", data)[0]
        self._current_tex.floatparams[0] = f

    def readSURF_TFP1(self, length):
        """TFP1 chunk."""
        if length!=4:
            raise LWOBError("Invalid TFP1 size (%d instead of 4)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TFP1 chunk")
        data = self.readNBytes(4)
        f = struct.unpack(">f", data)[0]
        self._current_tex.floatparams[1] = f

    def readSURF_TFP2(self, length):
        """TFP2 chunk."""
        if length!=4:
            raise LWOBError("Invalid TFP2 size (%d instead of 4)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TFP2 chunk")
        data = self.readNBytes(4)
        f = struct.unpack(">f", data)[0]
        self._current_tex.floatparams[2] = f

    def readSURF_TFP3(self, length):
        """TFP3 chunk."""
        if length!=4:
            raise LWOBError("Invalid TFP3 size (%d instead of 4)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TFP3 chunk")
        data = self.readNBytes(4)
        f = struct.unpack(">f", data)[0]
        self._current_tex.floatparams[3] = f

    def readSURF_TFP4(self, length):
        """TFP4 chunk."""
        if length!=4:
            raise LWOBError("Invalid TFP4 size (%d instead of 4)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TFP4 chunk")
        data = self.readNBytes(4)
        f = struct.unpack(">f", data)[0]
        self._current_tex.floatparams[4] = f

    def readSURF_TFP5(self, length):
        """TFP5 chunk."""
        if length!=4:
            raise LWOBError("Invalid TFP5 size (%d instead of 4)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TFP5 chunk")
        data = self.readNBytes(4)
        f = struct.unpack(">f", data)[0]
        self._current_tex.floatparams[5] = f

    def readSURF_TFP6(self, length):
        """TFP6 chunk."""
        if length!=4:
            raise LWOBError("Invalid TFP6 size (%d instead of 4)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TFP6 chunk")
        data = self.readNBytes(4)
        f = struct.unpack(">f", data)[0]
        self._current_tex.floatparams[6] = f

    def readSURF_TFP7(self, length):
        """TFP7 chunk."""
        if length!=4:
            raise LWOBError("Invalid TFP7 size (%d instead of 4)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TFP7 chunk")
        data = self.readNBytes(4)
        f = struct.unpack(">f", data)[0]
        self._current_tex.floatparams[7] = f

    def readSURF_TFP8(self, length):
        """TFP8 chunk."""
        if length!=4:
            raise LWOBError("Invalid TFP8 size (%d instead of 4)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TFP8 chunk")
        data = self.readNBytes(4)
        f = struct.unpack(">f", data)[0]
        self._current_tex.floatparams[8] = f

    def readSURF_TFP9(self, length):
        """TFP9 chunk."""
        if length!=4:
            raise LWOBError("Invalid TFP9 size (%d instead of 4)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TFP9 chunk")
        data = self.readNBytes(4)
        f = struct.unpack(">f", data)[0]
        self._current_tex.floatparams[9] = f

    def readSURF_TIP0(self, length):
        """TIP0 chunk."""
        if length!=2:
            raise LWOBError("Invalid TIP0 size (%d instead of 2)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TIP0 chunk")
        data = self.readNBytes(2)
        i = struct.unpack(">h", data)[0]
        self._current_tex.intparams[0] = i

    def readSURF_TIP1(self, length):
        """TIP1 chunk."""
        if length!=2:
            raise LWOBError("Invalid TIP1 size (%d instead of 2)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TIP1 chunk")
        data = self.readNBytes(2)
        i = struct.unpack(">h", data)[0]
        self._current_tex.intparams[1] = i

    def readSURF_TIP2(self, length):
        """TIP2 chunk."""
        if length!=2:
            raise LWOBError("Invalid TIP2 size (%d instead of 2)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TIP2 chunk")
        data = self.readNBytes(2)
        i = struct.unpack(">h", data)[0]
        self._current_tex.intparams[2] = i

    def readSURF_TIP3(self, length):
        """TIP3 chunk."""
        if length!=2:
            raise LWOBError("Invalid TIP3 size (%d instead of 2)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TIP3 chunk")
        data = self.readNBytes(2)
        i = struct.unpack(">h", data)[0]
        self._current_tex.intparams[3] = i

    def readSURF_TIP4(self, length):
        """TIP4 chunk."""
        if length!=2:
            raise LWOBError("Invalid TIP4 size (%d instead of 2)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TIP4 chunk")
        data = self.readNBytes(2)
        i = struct.unpack(">h", data)[0]
        self._current_tex.intparams[4] = i

    def readSURF_TIP5(self, length):
        """TIP5 chunk."""
        if length!=2:
            raise LWOBError("Invalid TIP5 size (%d instead of 2)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TIP5 chunk")
        data = self.readNBytes(2)
        i = struct.unpack(">h", data)[0]
        self._current_tex.intparams[5] = i

    def readSURF_TIP6(self, length):
        """TIP6 chunk."""
        if length!=2:
            raise LWOBError("Invalid TIP6 size (%d instead of 2)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TIP6 chunk")
        data = self.readNBytes(2)
        i = struct.unpack(">h", data)[0]
        self._current_tex.intparams[6] = i

    def readSURF_TIP7(self, length):
        """TIP7 chunk."""
        if length!=2:
            raise LWOBError("Invalid TIP7 size (%d instead of 2)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TIP7 chunk")
        data = self.readNBytes(2)
        i = struct.unpack(">h", data)[0]
        self._current_tex.intparams[7] = i

    def readSURF_TIP8(self, length):
        """TIP8 chunk."""
        if length!=2:
            raise LWOBError("Invalid TIP8 size (%d instead of 2)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TIP8 chunk")
        data = self.readNBytes(2)
        i = struct.unpack(">h", data)[0]
        self._current_tex.intparams[8] = i

    def readSURF_TIP9(self, length):
        """TIP9 chunk."""
        if length!=2:
            raise LWOBError("Invalid TIP9 size (%d instead of 2)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TIP9 chunk")
        data = self.readNBytes(2)
        i = struct.unpack(">h", data)[0]
        self._current_tex.intparams[9] = i

    def readSURF_TIMG(self, length):
        """TIMG chunk."""
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TIMG chunk")
        data = self.readNBytes(length)
        n = data.find("\000")
        if n!=-1:
            data = data[:n]

        self._current_tex.imagename = data

    def readSURF_TALP(self, length):
        """TALP chunk."""
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TALP chunk")
        data = self.readNBytes(length)
        n = data.find("\000")
        if n!=-1:
            data = data[:n]

        self._current_tex.alphaname = data

    def readSURF_TWRP(self, length):
        """TWRP chunk."""
        if length!=4:
            raise LWOBError("Invalid TWRP size (%d instead of 4)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TWRP chunk")
        data = self.readNBytes(4)
        w,h = struct.unpack(">HH", data)
        self._current_tex.widthwrap = w
        self._current_tex.heightwrap = h

    def readSURF_TAAS(self, length):
        """TAAS chunk."""
        if length!=4:
            raise LWOBError("Invalid TAAS size (%d instead of 4)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TAAS chunk")
        data = self.readNBytes(4)
        aa = struct.unpack(">f", data)[0]
        self._current_tex.antialiasingstrength = aa

    def readSURF_TOPC(self, length):
        """TOPC chunk."""
        if length!=4:
            raise LWOBError("Invalid TOPC size (%d instead of 4)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the TOPC chunk")
        data = self.readNBytes(4)
        op = struct.unpack(">f", data)[0]
        self._current_tex.opacity = op

    def readSURF_SHDR(self, length):
        """SHDR chunk."""
        if self._current_tex==None:
            raise LWOBError("Invalid position of the SHDR chunk")
        name = self.readNBytes(length)
        n = name.find("\000")
        if n!=-1:
            name = name[:n]

        self._current_tex.shader.append(name)

    def readSURF_SDAT(self, length):
        """SDAT chunk."""
        if self._current_tex==None:
            raise LWOBError("Invalid position of the SDAT chunk")
        data = self.readNBytes(length)
        self._current_tex.shaderdata.append(data)

    def readSURF_IMSQ(self, length):
        """IMSQ chunk."""
        if length!=6:
            raise LWOBError("Invalid IMSQ size (%d instead of 6)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the IMSQ chunk")
        data = self.readNBytes(6)
        offset,flags,looplen = struct.unpack(">HHH", data)
        self._current_tex.seq_offset = offset
        self._current_tex.seq_flags = flags
        self._current_tex.seq_looplength = looplen

    def readSURF_FLYR(self, length):
        """FLYR chunk."""
        if length!=8:
            raise LWOBError("Invalid FLYR size (%d instead of 8)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the FLYR chunk")
        data = self.readNBytes(8)
        b,e = struct.unpack(">II", data)
        self._current_tex.flyer_begin = b
        self._current_tex.flyer_end = e

    def readSURF_IMCC(self, length):
        """IMCC chunk."""
        if length!=6:
            raise LWOBError("Invalid IMCC size (%d instead of 6)"%length)
        if self._current_tex==None:
            raise LWOBError("Invalid position of the IMCC chunk")
        data = self.readNBytes(6)
        speed,low,high = struct.unpack(">HHH", data)
        self._current_tex.cycle_speed = speed
        self._current_tex.cycle_low = low
        self._current_tex.cycle_high = high


    #####

    def skipChunk(self, length):
        """Skip a chunk/sub chunk in the file (without reading it).

        Sets the file's position to the begin of the next chunk.
        length is the chunk length as it is stored in the file.
        The method takes care of adding the pad byte if length is odd.
        """
        if length%2==1:
            length += 1
        self._file.seek(length, 1)
        self._bytes_left -= length

    def readChunkHeader(self):
        """Read the head of the next chunk.

        Returns a tuple (tag, length).
        """
        s = self.readNBytes(8)
        tag = s[:4]
        length = struct.unpack(">I", s[4:])[0]
        return tag, length

    def readSubChunkHeader(self):
        """Read the head of the next sub chunk.

        Returns a tuple (tag, length).
        """
        s = self.readNBytes(6)
        tag = s[:4]
        length = struct.unpack(">H", s[4:])[0]
        return tag, length

    def readNBytes(self, n):
        """Read n bytes.

        Throws an exception if less than n bytes were read.
        """
        s = self._file.read(n)
        self._bytes_left -= len(s)
        if len(s)!=n:
            raise LWOBError("premature end of file")
        return s

    def _eofReached(self):
        """Return True when the end of the file has been reached.
        """
        return self._bytes_left<=1
            



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: mat4.py,v 1.2 2005/08/17 19:52:41 mbaas Exp $

import math, copy
from .vec3 import vec3 as _vec3
from .vec4 import vec4 as _vec4
from .mat3 import mat3 as _mat3


# [  0   1   2   3 ]
# [  4   5   6   7 ]
# [  8   9  10  11 ]
# [ 12  13  14  15 ]

# Comparison threshold
_epsilon = 1E-12


# mat4
class mat4:
    """Matrix class (4x4).

    This class represents a 4x4 matrix that can be used to store
    affine transformations.
    """

    def __init__(self, *args):
        "Constructor"

        # No arguments
        if len(args)==0:
            self.mlist = 16*[0.0]

        # 1 argument (list, scalar or mat4)
        elif len(args)==1:
            T = type(args[0])
            if T in [float, int]:
                f = float(args[0])
                self.mlist = [f,0.0,0.0,0.0,
                              0.0,f,0.0,0.0,
                              0.0,0.0,f,0.0,
                              0.0,0.0,0.0,f]
            # mat4
            elif isinstance(args[0], mat4):
                self.mlist = copy.copy(args[0].mlist)
            # String
            elif T is str:
                s=args[0].replace(","," ").replace("  "," ").strip().split(" ")
                self.mlist = list([float(x) for x in s])
            else:
                self.mlist = mat4(*args[0]).mlist

        # 4 arguments (sequences)
        elif len(args)==4:
            a,b,c,d=args
            self.mlist = [a[0], b[0], c[0], d[0],
                          a[1], b[1], c[1], d[1],
                          a[2], b[2], c[2], d[2],
                          a[3], b[3], c[3], d[3]]
            self.mlist = list([float(x) for x in self.mlist])

        # 16 arguments
        elif len(args)==16:
            self.mlist = list([float(x) for x in args])

        else:
            raise TypeError("mat4() arg can't be converted to mat4")

        # Check if there are really 16 elements in the list
        if len(self.mlist)!=16:
            raise TypeError("mat4(): Wrong number of matrix elements (%s instead of 16)"%(len(self.mlist)))

    def __repr__(self):
        return 'mat4(%s)'%(repr(self.mlist)[1:-1])

    def __str__(self):
        fmt="%9.4f"
        m11,m12,m13,m14,m21,m22,m23,m24,m31,m32,m33,m34,m41,m42,m43,m44 = self.mlist
        return ('['+fmt%m11+', '+fmt%m12+', '+fmt%m13+', '+fmt%m14+']\n'+
                '['+fmt%m21+', '+fmt%m22+', '+fmt%m23+', '+fmt%m24+']\n'+
                '['+fmt%m31+', '+fmt%m32+', '+fmt%m33+', '+fmt%m34+']\n'+
                '['+fmt%m41+', '+fmt%m42+', '+fmt%m43+', '+fmt%m44+']')

    def __eq__(self, other):
        """== operator"""
        global _epsilon
        if isinstance(other, mat4):
#            return self.mlist==other.mlist
            lst = list([a for a in zip(self.mlist, other.mlist) if abs(a[0]-a[1])>_epsilon])
            return len(lst)==0
        else:
            return False

    def __ne__(self, other):
        """!= operator"""
        return not (self==other)

    def __add__(self, other):
        """Matrix addition.

        >>> M=mat4(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16)
        >>> print M+M
        [   2.0000,    4.0000,    6.0000,    8.0000]
        [  10.0000,   12.0000,   14.0000,   16.0000]
        [  18.0000,   20.0000,   22.0000,   24.0000]
        [  26.0000,   28.0000,   30.0000,   32.0000]
        """
        if isinstance(other, mat4):
            return mat4(list(map(lambda x,y: x+y, self.mlist, other.mlist)))
        else:
            raise TypeError("unsupported operand type for +")

    def __sub__(self, other):
        """Matrix subtraction.

        >>> M=mat4(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16)
        >>> print M-M
        [   0.0000,    0.0000,    0.0000,    0.0000]
        [   0.0000,    0.0000,    0.0000,    0.0000]
        [   0.0000,    0.0000,    0.0000,    0.0000]
        [   0.0000,    0.0000,    0.0000,    0.0000]
        """
        if isinstance(other, mat4):
            return mat4(list(map(lambda x,y: x-y, self.mlist, other.mlist)))
        else:
            raise TypeError("unsupported operand type for -")

    def __mul__(self, other):
        """Multiplication.

        >>> M=mat4(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16)
        >>> print M*2.0
        [   2.0000,    4.0000,    6.0000,    8.0000]
        [  10.0000,   12.0000,   14.0000,   16.0000]
        [  18.0000,   20.0000,   22.0000,   24.0000]
        [  26.0000,   28.0000,   30.0000,   32.0000]
        >>> print 2.0*M
        [   2.0000,    4.0000,    6.0000,    8.0000]
        [  10.0000,   12.0000,   14.0000,   16.0000]
        [  18.0000,   20.0000,   22.0000,   24.0000]
        [  26.0000,   28.0000,   30.0000,   32.0000]
        >>> print M*M
        [  90.0000,  100.0000,  110.0000,  120.0000]
        [ 202.0000,  228.0000,  254.0000,  280.0000]
        [ 314.0000,  356.0000,  398.0000,  440.0000]
        [ 426.0000,  484.0000,  542.0000,  600.0000]
        >>> print M*_vec3(1,2,3)
        (0.1765, 0.4510, 0.7255)
        >>> print _vec3(1,2,3)*M
        (0.7083, 0.8056, 0.9028)
        """
        T = type(other)
        # mat4*scalar
        if T in [float, int]:
            return mat4(list(map(lambda x,other=other: x*other, self.mlist)))
        # mat4*vec3
        if isinstance(other, _vec3):
            m11,m12,m13,m14,m21,m22,m23,m24,m31,m32,m33,m34,m41,m42,m43,m44 = self.mlist
            w = float(m41*other.x + m42*other.y + m43*other.z + m44)
            return _vec3(m11*other.x + m12*other.y + m13*other.z + m14, 
                         m21*other.x + m22*other.y + m23*other.z + m24, 
                         m31*other.x + m32*other.y + m33*other.z + m34)/w
        # mat4*vec4
        if isinstance(other, _vec4):
            m11,m12,m13,m14,m21,m22,m23,m24,m31,m32,m33,m34,m41,m42,m43,m44 = self.mlist
            return _vec4(m11*other.x + m12*other.y + m13*other.z + m14*other.w, 
                         m21*other.x + m22*other.y + m23*other.z + m24*other.w, 
                         m31*other.x + m32*other.y + m33*other.z + m34*other.w,
                         m41*other.x + m42*other.y + m43*other.z + m44*other.w)
        # mat4*mat4
        if isinstance(other, mat4):
            m11,m12,m13,m14,m21,m22,m23,m24,m31,m32,m33,m34,m41,m42,m43,m44 = self.mlist
            n11,n12,n13,n14,n21,n22,n23,n24,n31,n32,n33,n34,n41,n42,n43,n44 = other.mlist
            return mat4( m11*n11+m12*n21+m13*n31+m14*n41,
                         m11*n12+m12*n22+m13*n32+m14*n42,
                         m11*n13+m12*n23+m13*n33+m14*n43,
                         m11*n14+m12*n24+m13*n34+m14*n44,

                         m21*n11+m22*n21+m23*n31+m24*n41,
                         m21*n12+m22*n22+m23*n32+m24*n42,
                         m21*n13+m22*n23+m23*n33+m24*n43,
                         m21*n14+m22*n24+m23*n34+m24*n44,

                         m31*n11+m32*n21+m33*n31+m34*n41,
                         m31*n12+m32*n22+m33*n32+m34*n42,
                         m31*n13+m32*n23+m33*n33+m34*n43,
                         m31*n14+m32*n24+m33*n34+m34*n44,

                         m41*n11+m42*n21+m43*n31+m44*n41,
                         m41*n12+m42*n22+m43*n32+m44*n42,
                         m41*n13+m42*n23+m43*n33+m44*n43,
                         m41*n14+m42*n24+m43*n34+m44*n44)
        # unsupported
        else:
            raise TypeError("unsupported operand type for *")

    def __rmul__(self, other):
        T = type(other)
        # scalar*mat4
        if T in [float, int]:
            return mat4(list(map(lambda x,other=other: other*x, self.mlist)))
        # vec4*mat4
        if isinstance(other, _vec4):
            m11,m12,m13,m14,m21,m22,m23,m24,m31,m32,m33,m34,m41,m42,m43,m44 = self.mlist
            return _vec4(other.x*m11 + other.y*m21 + other.z*m31 + other.w*m41, 
                         other.x*m12 + other.y*m22 + other.z*m32 + other.w*m42,
                         other.x*m13 + other.y*m23 + other.z*m33 + other.w*m43,
                         other.x*m14 + other.y*m24 + other.z*m34 + other.w*m44)
        # vec3*mat4
        if isinstance(other, _vec3):
            m11,m12,m13,m14,m21,m22,m23,m24,m31,m32,m33,m34,m41,m42,m43,m44 = self.mlist
            w = float(other.x*m14 + other.y*m24 + other.z*m34 + m44)
            return _vec3(other.x*m11 + other.y*m21 + other.z*m31 + m41, 
                         other.x*m12 + other.y*m22 + other.z*m32 + m42,
                         other.x*m13 + other.y*m23 + other.z*m33 + m43)/w
        # mat4*mat4
        if isinstance(other, mat4):
            return self.__mul__(other)
        # unsupported
        else:
            raise TypeError("unsupported operand type for *")

    def __truediv__(self, other):
        """Division
        
        >>> M=mat4(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16)
        >>> print M/2.0
        [   0.5000,    1.0000,    1.5000,    2.0000]
        [   2.5000,    3.0000,    3.5000,    4.0000]
        [   4.5000,    5.0000,    5.5000,    6.0000]
        [   6.5000,    7.0000,    7.5000,    8.0000]
        """
        T = type(other)
        # mat4/scalar
        if T in [float, int]:
            return mat4(list(map(lambda x,other=other: x/other, self.mlist)))
        # unsupported
        else:
            raise TypeError("unsupported operand type for /")

    def __mod__(self, other):
        """Modulo.

        >>> M=mat4(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16)
        >>> print M%5.0
        [   1.0000,    2.0000,    3.0000,    4.0000]
        [   0.0000,    1.0000,    2.0000,    3.0000]
        [   4.0000,    0.0000,    1.0000,    2.0000]
        [   3.0000,    4.0000,    0.0000,    1.0000]
        """
        T = type(other)
        # mat4%scalar
        if T in [float, int]:
            return mat4(list(map(lambda x,other=other: x%other, self.mlist)))
        # mat4%mat4
        if isinstance(other, mat4):
            return mat4(list(map(lambda a,b: a%b, self.mlist, other.mlist)))
        # unsupported
        else:
            raise TypeError("unsupported operand type for %")

    def __neg__(self):
        """Negation.

        >>> M=mat4(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16)
        >>> print -M
        [  -1.0000,   -2.0000,   -3.0000,   -4.0000]
        [  -5.0000,   -6.0000,   -7.0000,   -8.0000]
        [  -9.0000,  -10.0000,  -11.0000,  -12.0000]
        [ -13.0000,  -14.0000,  -15.0000,  -16.0000]
        """
        return mat4([-x for x in self.mlist])

    def __pos__(self):
        """
        >>> M=mat4(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16)
        >>> print +M
        [   1.0000,    2.0000,    3.0000,    4.0000]
        [   5.0000,    6.0000,    7.0000,    8.0000]
        [   9.0000,   10.0000,   11.0000,   12.0000]
        [  13.0000,   14.0000,   15.0000,   16.0000]
        """
        return mat4([+x for x in self.mlist])


    def __len__(self):
        return 4

    def __getitem__(self, key):
        if type(key) is int:
            if key<0 or key>3:
                raise IndexError("index out of range")
            m=self.mlist
            if   key==0: return _vec4(m[0],m[4],m[8],m[12])
            elif key==1: return _vec4(m[1],m[5],m[9],m[13])
            elif key==2: return _vec4(m[2],m[6],m[10],m[14])
            elif key==3: return _vec4(m[3],m[7],m[11],m[15])
        elif type(key) is tuple:
            i,j=key
            if i<0 or i>3 or j<0 or j>3:
                raise IndexError("index out of range")
            return self.mlist[i*4+j]
        else:
            raise TypeError("index must be integer or 2-tuple")

    def __setitem__(self, key, value):
        if type(key) is int:
            if key<0 or key>3:
                raise IndexError("index out of range")
            m=self.mlist
            value = [float(x) for x in value]
            if   key==0: m[0],m[4],m[8],m[12]=value
            elif key==1: m[1],m[5],m[9],m[13]=value
            elif key==2: m[2],m[6],m[10],m[14]=value
            elif key==3: m[3],m[7],m[11],m[15]=value
        elif type(key) is tuple:
            i,j=key
            if i<0 or i>3 or j<0 or j>3:
                raise IndexError("index out of range")
            self.mlist[i*4+j] = float(value)
        else:
            raise TypeError("index must be integer or 2-tuple")

    def getRow(self, index):
        """Return row (as vec4)."""
        if type(index)==int:
            m=self.mlist
            if   index==0: return _vec4(m[0], m[1], m[2], m[3])
            elif index==1: return _vec4(m[4], m[5], m[6], m[7])
            elif index==2: return _vec4(m[8], m[9], m[10], m[11])
            elif index==3: return _vec4(m[12], m[13], m[14], m[15])
            else:
                raise IndexError("index out of range")
        else:
            raise TypeError("index must be integer or 2-tuple")

    def setRow(self, index, value):
        """Set row."""
        if type(index)==int:
            m=self.mlist
            value = [float(x) for x in value]
            if   index==0: m[0],m[1],m[2],m[3] = value
            elif index==1: m[4],m[5],m[6],m[7] = value
            elif index==2: m[8],m[9],m[10],m[11] = value
            elif index==3: m[12],m[13],m[14],m[15] = value
            else:
                raise IndexError("index out of range")
        else:
            raise TypeError("index must be integer or 2-tuple")

    def getColumn(self, index):
        """Return column (as vec4)."""
        if type(index)==int:
            m=self.mlist
            if   index==0: return _vec4(m[0], m[4], m[8], m[12])
            elif index==1: return _vec4(m[1], m[5], m[9], m[13])
            elif index==2: return _vec4(m[2], m[6], m[10], m[14])
            elif index==3: return _vec4(m[3], m[7], m[11], m[15])
            else:
                raise IndexError("index out of range")
        else:
            raise TypeError("index must be integer or 2-tuple")

    def setColumn(self, index, value):
        """Set column."""
        if type(index)==int:
            m = self.mlist
            value = [float(x) for x in value]
            if   index==0: m[0],m[4],m[8],m[12] = value
            elif index==1: m[1],m[5],m[9],m[13] = value
            elif index==2: m[2],m[6],m[10],m[14] = value
            elif index==3: m[3],m[7],m[11],m[15] = value
            else:
                raise IndexError("index out of range")
        else:
            raise TypeError("index must be integer or 2-tuple")

    def getDiag(self):
        """Return the diagonal."""
        return _vec4(self.mlist[0], self.mlist[5], self.mlist[10], self.mlist[15])

    def setDiag(self, value):
        """Set diagonal."""
        a,b,c,d = value
        self.mlist[0] = float(a)
        self.mlist[5] = float(b)
        self.mlist[10] = float(c)
        self.mlist[15] = float(d)

    def toList(self, rowmajor=0):
        """Return a list containing the matrix elements.

        By default the list is in column-major order (which can directly be
        used in OpenGL or RenderMan). If you set the optional argument
        rowmajor to 1, you'll get the list in row-major order.

        >>> M=mat4(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16)
        >>> print M.toList()
        [1, 5, 9, 13, 2, 6, 10, 14, 3, 7, 11, 15, 4, 8, 12, 16]
        >>> print M.toList(rowmajor=1)
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16]
        """
        if rowmajor:
            return copy.copy(self.mlist)
        else:
            return self.transpose().mlist
            

    def identity():
        """Return identity matrix.

        >>> print mat4().identity()
        [   1.0000,    0.0000,    0.0000,    0.0000]
        [   0.0000,    1.0000,    0.0000,    0.0000]
        [   0.0000,    0.0000,    1.0000,    0.0000]
        [   0.0000,    0.0000,    0.0000,    1.0000]
        """
        return mat4(1.0, 0.0, 0.0, 0.0,
                    0.0, 1.0, 0.0, 0.0,
                    0.0, 0.0, 1.0, 0.0,
                    0.0, 0.0, 0.0, 1.0)
    identity = staticmethod(identity)

    def transpose(self):
        """Transpose matrix.
        
        >>> M=mat4(1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16)
        >>> print M.transpose()
        [   1.0000,    5.0000,    9.0000,   13.0000]
        [   2.0000,    6.0000,   10.0000,   14.0000]
        [   3.0000,    7.0000,   11.0000,   15.0000]
        [   4.0000,    8.0000,   12.0000,   16.0000]
        """
        m11,m12,m13,m14,m21,m22,m23,m24,m31,m32,m33,m34,m41,m42,m43,m44 = self.mlist
        return mat4(m11,m21,m31,m41,
                    m12,m22,m32,m42,
                    m13,m23,m33,m43,
                    m14,m24,m34,m44)

    def determinant(self):
        """Return determinant.
        
        >>> M=mat4(2.0,0,0,0, 0,2.0,0,0, 0,0,2.0,0, 0,0,0,2.0)
        >>> print M.determinant()
        16.0
        """
        m11,m12,m13,m14,m21,m22,m23,m24,m31,m32,m33,m34,m41,m42,m43,m44 = self.mlist

        return m11*m22*m33*m44 \
               -m11*m22*m34*m43 \
               +m11*m23*m34*m42 \
               -m11*m23*m32*m44 \
               +m11*m24*m32*m43 \
               -m11*m24*m33*m42 \
               -m12*m23*m34*m41 \
               +m12*m23*m31*m44 \
               -m12*m24*m31*m43 \
               +m12*m24*m33*m41 \
               -m12*m21*m33*m44 \
               +m12*m21*m34*m43 \
               +m13*m24*m31*m42 \
               -m13*m24*m32*m41 \
               +m13*m21*m32*m44 \
               -m13*m21*m34*m42 \
               +m13*m22*m34*m41 \
               -m13*m22*m31*m44 \
               -m14*m21*m32*m43 \
               +m14*m21*m33*m42 \
               -m14*m22*m33*m41 \
               +m14*m22*m31*m43 \
               -m14*m23*m31*m42 \
               +m14*m23*m32*m41
    

    def _submat(self, i,j):
        M=_mat3()
        for k in range(3):
            for l in range(3):
                t=(k,l)
                if k>=i:
                    t=(k+1,t[1])
                if l>=j:
                    t=(t[0],l+1)
                M[k,l] = self[t]
        return M
        
    def inverse(self):
        """Return inverse matrix.

        >>> M=mat4(0,-2.0,0,0, 2.0,0,0,0, 0,0,2,0, 0,0,0,2)
        >>> print M.inverse()
        [   0.0000,    0.5000,    0.0000,    0.0000]
        [  -0.5000,    0.0000,    0.0000,    0.0000]
        [   0.0000,    0.0000,    0.5000,    0.0000]
        [   0.0000,    0.0000,    0.0000,    0.5000]
        """
        
        Mi=mat4()
        d=self.determinant()
        for i in range(4):
            for j in range(4):
                sign=1-((i+j)%2)*2
                m3=self._submat(i,j)
                Mi[j,i]=sign*m3.determinant()/d
        return Mi

    def translation(t):
        """Return translation matrix."""
        tx,ty,tz = t
        return mat4(1.0, 0.0, 0.0, tx,
                    0.0, 1.0, 0.0, ty,
                    0.0, 0.0, 1.0, tz,
                    0.0, 0.0, 0.0, 1.0)
    translation = staticmethod(translation)

    def scaling(s):
        """Return scaling matrix."""
        sx,sy,sz = s
        return mat4(sx, 0.0, 0.0, 0.0,
                    0.0, sy, 0.0, 0.0,
                    0.0, 0.0, sz, 0.0,
                    0.0, 0.0, 0.0, 1.0)
    scaling = staticmethod(scaling)

    def rotation(angle, axis):
        """Return rotation matrix.

        angle must be given in radians. axis should be of type vec3.
        """
        axis = _vec3(axis)

        sqr_a = axis.x*axis.x
        sqr_b = axis.y*axis.y
        sqr_c = axis.z*axis.z
        len2  = sqr_a+sqr_b+sqr_c

        k2    = math.cos(angle)
        k1    = (1.0-k2)/len2
        k3    = math.sin(angle)/math.sqrt(len2)
        k1ab  = k1*axis.x*axis.y
        k1ac  = k1*axis.x*axis.z
        k1bc  = k1*axis.y*axis.z
        k3a   = k3*axis.x
        k3b   = k3*axis.y
        k3c   = k3*axis.z

        return mat4( k1*sqr_a+k2, k1ab-k3c, k1ac+k3b, 0.0,
                     k1ab+k3c, k1*sqr_b+k2, k1bc-k3a, 0.0,
                     k1ac-k3b, k1bc+k3a, k1*sqr_c+k2, 0.0,
                     0.0, 0.0, 0.0, 1.0)
    rotation = staticmethod(rotation)

    def translate(self, t):
        """Concatenate a translation."""
        tx = float(t[0])
        ty = float(t[1])
        tz = float(t[2])
        m11,m12,m13,m14,m21,m22,m23,m24,m31,m32,m33,m34,m41,m42,m43,m44 = self.mlist
        self.mlist[3]  = m11*tx + m12*ty + m13*tz + m14
        self.mlist[7]  = m21*tx + m22*ty + m23*tz + m24
        self.mlist[11] = m31*tx + m32*ty + m33*tz + m34
        self.mlist[15] = m41*tx + m42*ty + m43*tz + m44
        return self

    def scale(self, s):
        """Concatenate a scaling."""
        sx = float(s[0])
        sy = float(s[1])
        sz = float(s[2])
        self.mlist[0]  *= sx
        self.mlist[1]  *= sy
        self.mlist[2]  *= sz
        self.mlist[4]  *= sx
        self.mlist[5]  *= sy
        self.mlist[6]  *= sz
        self.mlist[8]  *= sx
        self.mlist[9]  *= sy
        self.mlist[10] *= sz
        self.mlist[12] *= sx
        self.mlist[13] *= sy
        self.mlist[14] *= sz
        return self

    def rotate(self, angle, axis):
        """Concatenate a rotation.

        angle must be given in radians. axis should be of type vec3.
        """
        R=self.rotation(angle, axis)
        self.mlist = (self*R).mlist
        return self


    def frustum(left, right, bottom, top, near, far):
        """Set a perspective transformation.
        
        This method is equivalent to the OpenGL command glFrustum().
        """
        return mat4( (2.0*near)/(right-left), 0.0, float(right+left)/(right-left), 0.0,
                     0.0, (2.0*near)/(top-bottom), float(top+bottom)/(top-bottom), 0.0,
                     0.0, 0.0, -float(far+near)/(far-near), -(2.0*far*near)/(far-near),
                     0.0, 0.0, -1.0, 0.0)
    frustum = staticmethod(frustum)
    
    def perspective(fovy, aspect, near, far):
        """ Set a perspective transformation.

        This method is equivalent to the OpenGL utility command
        gluPerspective().
        """
        top    = near * math.tan(fovy * math.pi / 360.0)
        bottom = -top
        left   = bottom * aspect
        right  = top * aspect

        return mat4.frustum(left, right, bottom, top, near, far)
    perspective = staticmethod(perspective)

    def orthographic(left, right, bottom, top, near, far):
        """Returns a matrix that represents an orthographic transformation.

        This method is equivalent to the OpenGL command glOrtho().
        """
        global _epsilon
        
        r_l = float(right-left)
        t_b = float(top-bottom)
        f_n = float(far-near)
  
        if r_l<=_epsilon:
            raise ValueError("right-value must be greater than left-value")
        if t_b<=_epsilon:
            raise ValueError("top-value must be greater than bottom-value")
        if f_n<=_epsilon:
            raise ValueError("far-value must be greater than near-value")

        return mat4(2.0/r_l, 0.0, 0.0, -(right+left)/r_l,
                    0.0, 2.0/t_b, 0.0, -(top+bottom)/t_b,
                    0.0, 0.0, -2.0/f_n, -(far+near)/f_n,
                    0.0, 0.0, 0.0, 1.0)
    orthographic = staticmethod(orthographic)

    def lookAt(pos, target, up=_vec3(0,0,1)):
        """Look from pos to target.

        The resulting transformation moves the origin to pos and
        rotates so that the z-axis points to target. The y-axis is
        as close as possible to the up vector.
        """
        pos = _vec3(pos)
        target = _vec3(target)
        up = _vec3(up)
        dir = (target - pos).normalize()
        up  = up.normalize()
        up -= (up * dir) * dir
        try:
            up  = up.normalize()
        except:
            # We're looking along the up direction, so choose
            # an arbitrary direction that is perpendicular to dir
            # as new up.
            up = dir.ortho()

        right = up.cross(dir).normalize()

        return mat4(right.x, up.x, dir.x, pos.x,
                    right.y, up.y, dir.y, pos.y,
                    right.z, up.z, dir.z, pos.z,
                    0.0, 0.0, 0.0, 1.0)
    lookAt = staticmethod(lookAt)

    def ortho(self):
        """Return a matrix with orthogonal base vectors.

        Makes the x-, y- and z-axis orthogonal.
        The fourth column and row remain untouched.
        """

        m11,m12,m13,m14,m21,m22,m23,m24,m31,m32,m33,m34,m41,m42,m43,m44 = self.mlist

        x = _vec3(m11, m21, m31)
        y = _vec3(m12, m22, m32)
        z = _vec3(m13, m23, m33)

        xl = x.length()
        xl*=xl
        y = y - ((x*y)/xl)*x
        z = z - ((x*z)/xl)*x

        yl = y.length()
        yl*=yl
        z = z - ((y*z)/yl)*y

        return mat4( x.x, y.x, z.x, m14,
                     x.y, y.y, z.y, m24,
                     x.z, y.z, z.z, m34,
                     m41, m42, m43, m44)

    def decompose(self):
        """Decomposes the matrix into a translation, rotation and scaling part.

        Returns a tuple (translation, rotation, scaling). The 
        translation and scaling parts are given as vec3's, the rotation
        is still given as a mat4.
        """
        try:
            dummy = self.ortho()
        except ZeroDivisionError:
            return (_vec3(0), mat4(1.0), _vec3(0))

        dummy.setRow(3,_vec4(0.0, 0.0, 0.0, 1.0))
        dummy.setColumn(3,_vec4(0.0, 0.0, 0.0, 1.0))

        x = dummy.getColumn(0)
        y = dummy.getColumn(1)
        z = dummy.getColumn(2)
        xl = x.length()
        yl = y.length()
        zl = z.length()
        scale = _vec3(xl,yl,zl)
        
        x/=xl
        y/=yl
        z/=zl
        dummy.setColumn(0,x)
        dummy.setColumn(1,y)
        dummy.setColumn(2,z)
        if dummy.determinant()<0.0:
            dummy.setColumn(0,-x)
            scale.x=-scale.x

        return (_vec3(self.mlist[3], self.mlist[7], self.mlist[11]),
                dummy,
                scale)

    def getMat3(self):
        """Convert to mat3 by discarding 4th row and column.
        """
        m11,m12,m13,m14,m21,m22,m23,m24,m31,m32,m33,m34,m41,m42,m43,m44 = self.mlist
        return _mat3(m11,m12,m13,
                     m21,m22,m23,
                     m31,m32,m33)

    def setMat3(self, m3):
        """Sets the first three columns and rows to the values in m3.
        """
        self.mlist[0:3] = m3.mlist[0:3]
        self.mlist[4:7] = m3.mlist[3:6]
        self.mlist[8:11] = m3.mlist[6:9]

######################################################################

def _test():
    import doctest, mat4
    failed, total = doctest.testmod(mat4)
    print(("%d/%d failed" % (failed, total)))

if __name__=="__main__":

    _test()


